(**************************************************************************)
(*                                                                        *)
(*  SSH Connection Protocol                                               *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Channels;

        (********************************************************)
        (*                                                      *)
        (*               Connection layer of SSH                *)
        (*       This is the client end of the connection       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            24 November 2021                *)
        (*  Last edited:        21 September 2022               *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, CAST;

IMPORT Strings;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  SendPacket;

FROM PacketQueues IMPORT
    (* type *)  PacketQ,
    (* proc *)  GetPacket;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, BSDelete, GetCharString, GetCard, AddCard,
                AddByte, AddByteString, AddCharString;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, EVAL;

FROM Consts IMPORT
    (* const*)  SSH_MSG_CHANNEL_OPEN, SSH_MSG_CHANNEL_OPEN_CONFIRMATION, SSH_MSG_CHANNEL_OPEN_FAILURE,
                SSH_MSG_CHANNEL_DATA, SSH_MSG_CHANNEL_CLOSE, SSH_MSG_CHANNEL_REQUEST, SSH_MSG_TERMINATE;

(************************************************************************)

CONST
    SSH_OPEN_ADMINISTRATIVELY_PROHIBITED    = 1;
    SSH_OPEN_CONNECT_FAILED                 = 2;
    SSH_OPEN_UNKNOWN_CHANNEL_TYPE           = 3;
    SSH_OPEN_RESOURCE_SHORTAGE              = 4;

(************************************************************************)
(*                  THE POOL OF AVAILABLE CHANNELS                      *)
(************************************************************************)

CONST
    MaxChannelNo = 128;
    NotAChannel = MaxChannelNo + 1;

TYPE
    (* prevfree and nextfree form a linked list of available channels.  *)
    (* peer=0 means a closed channel, even if it hasn't yet been        *)
    (* deallocated.                                                     *)

    ChannelInfoType =   RECORD
                            prevfree, nextfree: CARDINAL;
                            peer: CARDINAL;
                            port: IOPort;
                            Q: PacketQ;
                        END (*RECORD*);

VAR
    (* Head and tail of list of available channels. *)

    FreeChans:  RECORD
                    head, tail: CARDINAL;
                END (*RECORD*);

    (* The status of all channels. *)

    ChannelStatus: ARRAY [1..MaxChannelNo] OF ChannelInfoType;

(************************************************************************)

PROCEDURE ClearChannelData;

    (* Initialises the ChannelStatus array.  *)

    VAR chan: CARDINAL;

    BEGIN
        FOR chan := 1 TO MaxChannelNo DO
            WITH ChannelStatus[chan] DO
                prevfree := chan - 1;
                nextfree := chan + 1;
                peer := NotAChannel;
                port := NIL;
                Q := NIL;
            END (*WITH*);
        END (*FOR*);
        ChannelStatus[MaxChannelNo].nextfree := 0;
        FreeChans.head := 1;
        FreeChans.tail := MaxChannelNo;
    END ClearChannelData;

(************************************************************************)

PROCEDURE ObtainChannel(): CARDINAL;

    (* Takes a channel from the free list, or returns 0 if no channel available. *)

    VAR chan: CARDINAL;

    BEGIN
        chan := FreeChans.head;
        IF chan <> 0 THEN
            FreeChans.head := ChannelStatus[chan].nextfree;
            ChannelStatus[chan].peer := MAX(CARDINAL);  (* to mark it used *)
        END (*IF*);
        IF FreeChans.head = 0 THEN
            FreeChans.tail := 0;
        END (*IF*);
        ChannelStatus[chan].nextfree := 0;
        ChannelStatus[chan].prevfree := 0;
        RETURN chan;
    END ObtainChannel;

(************************************************************************)

PROCEDURE ReleaseChannel (chan: CARDINAL);

    (* Returns a channel to the free list. *)

    VAR prev, next: CARDINAL;

    BEGIN
        ChannelStatus[chan].peer := NotAChannel;
        ChannelStatus[chan].port := NIL;
        ChannelStatus[chan].Q := NIL;

        (* Find the next free channel. *)

        next := chan + 1;
        LOOP
            IF next > MaxChannelNo THEN
                next := 1;
            END (*IF*);
            IF (next = chan) OR (ChannelStatus[next].peer = NotAChannel) THEN
                EXIT (*LOOP*);
            END (*IF*);
            INC (next);
        END (*LOOP*);

        IF next = chan THEN
            ChannelStatus[chan].nextfree := 0;
            ChannelStatus[chan].prevfree := 0;
            FreeChans.head := chan;
            FreeChans.tail := chan;
        ELSE
            prev := ChannelStatus[next].prevfree;
            ChannelStatus[chan].nextfree := next;
            ChannelStatus[chan].prevfree := prev;
            ChannelStatus[prev].nextfree := chan;
            ChannelStatus[next].prevfree := chan;
        END (*IF*);

    END ReleaseChannel;

(************************************************************************)
(*                   OPENING AND CLOSING A CHANNEL                      *)
(************************************************************************)

PROCEDURE ClntOpenAChannel (port: IOPort;  Q: PacketQ): CARDINAL;

    (* Opens a new channel, including asking the server to open it at   *)
    (* the other end.  Returns out channel number, or 0 for failure.    *)

    CONST
        ReportSuccess = FALSE;
        channeltype = "session";
        windowsize = 8192;
        maxpacketsize = 65536;

    VAR pkt, reply: ByteStr;
        ourchannelno: CARDINAL;
        k, pos, peerwsz, peermaxpacksz, reason: CARDINAL;
        opcode: CARD8;
        chantype: ARRAY [0..31] OF CHAR;
        message: ARRAY [0..127] OF CHAR;
        success: BOOLEAN;

    BEGIN
        ourchannelno := ObtainChannel();
        success := ourchannelno <> 0;
        IF success THEN
            ChannelStatus[ourchannelno].port := port;
            ChannelStatus[ourchannelno].Q := Q;
        END (*IF*);

        (* Now ask the server to open its end of the channel. *)

        IF success THEN
            MakeBS (pkt, 5 + LENGTH(channeltype) + 12);
            pkt.data^[0] := SSH_MSG_CHANNEL_OPEN;
            k := 1;
            AddCharString (pkt, k, channeltype);
            AddCard (pkt, k, ourchannelno);
            AddCard (pkt, k, windowsize);
            AddCard (pkt, k, maxpacketsize);
            success := SendPacket (port, pkt);
        END (*IF*);

        IF NOT success THEN
            IF ourchannelno <> 0 THEN
                ReleaseChannel (ourchannelno);
            END (*IF*);
            RETURN 0;
        END (*IF*);

        (* Check the server's reply. *)

        reply := GetPacket(Q);
        opcode := reply.data^[0];
        pos := 1;
        IF opcode = SSH_MSG_CHANNEL_OPEN_CONFIRMATION THEN
            k := GetCard (reply, pos);
            ChannelStatus[ourchannelno].peer := GetCard (reply, pos);
            peerwsz := GetCard (reply, pos);
            peermaxpacksz := GetCard (reply, pos);
            IF k <> ourchannelno THEN
                WriteString ("Channel number mismatch in OpenChannel");
                WriteLn;
            ELSE
                IF ReportSuccess THEN
                    Strings.Assign ("Server reports channel type ", message);
                    Strings.Append (chantype, message);
                    WriteString (message);  WriteLn;
                    Strings.Assign ("  peer channel # = ", message);
                    AppendCard (ChannelStatus[ourchannelno].peer, message);
                    WriteString (message);  WriteLn;
                    Strings.Assign ("  peer window size = ", message);
                    AppendCard (peerwsz, message);
                    WriteString (message);  WriteLn;
                    Strings.Assign ("  peer max packet size = ", message);
                    AppendCard (peermaxpacksz, message);
                    WriteString (message);  WriteLn;
                END (*IF*);
            END (*IF*);
        ELSIF opcode = SSH_MSG_CHANNEL_OPEN_FAILURE THEN
            k := GetCard (reply, pos);
            reason := GetCard (reply, pos);
            GetCharString (reply, pos, chantype);
            Strings.Assign ("channel open error ", message);
            AppendCard (reason, message);
            Strings.Append (": ", message);
            Strings.Append (chantype, message);
            WriteString (message);  WriteLn;
            ReleaseChannel (ourchannelno);
            ourchannelno := 0;
        ELSE
            Strings.Assign ("Unexpected message type ", message);
            AppendCard (opcode, message);
            WriteString (message);  WriteLn;
        END (*IF*);

        DiscardBS (reply);
        RETURN ourchannelno;

    END ClntOpenAChannel;

(************************************************************************)

PROCEDURE CloseChannel (chan: CARDINAL);

    (* Closes the channel. *)

    VAR pos: CARDINAL;
        pkt, reply: ByteStr;
        success: BOOLEAN;

    BEGIN
        (* Send a close message to the server,    *)

        success := ChannelStatus[chan].peer <> NotAChannel;
        IF success THEN
            MakeBS (pkt, 5);
            pkt.data^[0] := SSH_MSG_CHANNEL_CLOSE;
            pos := 1;
            AddCard (pkt, pos, ChannelStatus[chan].peer);
            success := SendPacket (ChannelStatus[chan].port, pkt);
        END (*IF*);

        (* Wait for a reply.  *)

        IF success THEN
            reply := GetPacket(ChannelStatus[chan].Q);
            DiscardBS (reply);
        END (*IF*);

        (* Close our end of the channel. *)

        ReleaseChannel (chan);

    END CloseChannel;

(************************************************************************)

PROCEDURE IsClosedChannel (chan: CARDINAL): BOOLEAN;

    (* Returns TRUE iff this is no longer an active channel. *)

    BEGIN
        RETURN ChannelStatus[chan].peer = 0;
    END IsClosedChannel;

(************************************************************************)
(*                          CHANNEL REQUEST                             *)
(************************************************************************)

PROCEDURE RequestSubsystem (ourchan: CARDINAL;  subsystem: ARRAY OF CHAR);

    (* Sends an SSH_MSG_CHANNEL_REQUEST message.  *)

    CONST NILQ = CAST (PacketQ, NIL);

    VAR pos: CARDINAL;
        pkt: ByteStr;

    BEGIN
        MakeBS (pkt, 14 + LENGTH("subsystem") + LENGTH(subsystem));
        pkt.data^[0] := SSH_MSG_CHANNEL_REQUEST;
        pos := 1;
        AddCard (pkt, pos, ChannelStatus[ourchan].peer);
        AddCharString (pkt, pos, "subsystem");
        AddByte (pkt, pos, 0);
        AddCharString (pkt, pos, subsystem);
        EVAL (SendPacket (ChannelStatus[ourchan].port, pkt));

        (* No reply is expected.  *)

    END RequestSubsystem;

(************************************************************************)
(*                    RECEIVING PACKET FROM SERVER                      *)
(************************************************************************)

PROCEDURE ReceivePacket (chan: CARDINAL;  VAR (*OUT*) result: ByteStr): BOOLEAN;

    (* Gets a packet from the server.  Returns FALSE on connection lost. *)

    BEGIN
        result := GetPacket(ChannelStatus[chan].Q);
        IF result.data^[0] = SSH_MSG_TERMINATE THEN
            WriteString ("Channel received TERMINATE message");
            WriteLn;
            ChannelStatus[chan].peer := 0;
            DiscardBS (result);
            RETURN FALSE;
        ELSE
            (* First 9 bytes are SSH overhead. *)

            BSDelete (result, 0, 9);
        END (*IF*);
        RETURN TRUE;

    END ReceivePacket;

(************************************************************************)
(*                           DATA TRANSFER                              *)
(************************************************************************)

PROCEDURE SendData (channel: CARDINAL;  pkt: ByteStr): BOOLEAN;

    (* Sends a data packet to peer. *)

    VAR pos: CARDINAL;
        tosend: ByteStr;
        success: BOOLEAN;

    BEGIN
        MakeBS (tosend, 9 + pkt.size);
        tosend.data^[0] := SSH_MSG_CHANNEL_DATA;
        pos := 1;
        AddCard (tosend, pos, ChannelStatus[channel].peer);
        AddByteString (tosend, pos, pkt.data^, pkt.size);
        DiscardBS (pkt);
        success := SendPacket (ChannelStatus[channel].port, tosend);
        IF NOT success THEN
            ChannelStatus[channel].peer := 0;
        END (*IF*);
        RETURN success;
    END SendData;

(************************************************************************)

BEGIN
    ClearChannelData;
END Channels.

