(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE ClientLogin;

        (********************************************************)
        (*                                                      *)
        (*               SSH client: logging in                 *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            12 June 2021                    *)
        (*  Last edited:        13 January 2022                *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (*  We only support the username/password form of login *)
        (*  Anything else seems inappropriate for FTP.          *)
        (*                                                      *)
        (********************************************************)


IMPORT PacketQueues, Services;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  SendPacket;

FROM Consts IMPORT
    (* const*)  SSH_MSG_SERVICE_ACCEPT, SSH_MSG_USERAUTH_REQUEST, SSH_MSG_USERAUTH_SUCCESS;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, AddCharString;

FROM FileOps IMPORT
    (* proc *)  ReadLine, ReadKbdLine, FWriteString, FWriteLn;

(********************************************************************************)

CONST Nul = CHR(0);

(********************************************************************************)
(*                          THE ACTUAL LOGIN OPERATION                          *)
(********************************************************************************)

PROCEDURE TryLogin (port: IOPort;  replyQ: PacketQueues.PacketQ;
                           username, password, methodname: ARRAY OF CHAR;
                           VAR (*OUT*) connectionlost: BOOLEAN): BOOLEAN;

    (* Sends a login request, checks the response.  *)

    CONST
        servicename = "ssh-connection";

    VAR pkt, reply: ByteStr;
        k, L0, L3: CARDINAL;
        success: BOOLEAN;

    BEGIN
        L0 := LENGTH(username);  L3 := LENGTH(password);
        MakeBS (pkt, 5 + L0 + 4 + LENGTH(servicename) + 4 + LENGTH(methodname) + 5 + L3);
        pkt.data^[0] := SSH_MSG_USERAUTH_REQUEST;
        k := 1;
        AddCharString (pkt, k, username);
        AddCharString (pkt, k, servicename);
        AddCharString (pkt, k, methodname);
        pkt.data^[k] := 0;  INC (k);
        AddCharString (pkt, k, password);
        success := SendPacket (port, pkt);
        connectionlost := NOT success;

        IF success THEN
            reply := PacketQueues.GetPacket (replyQ);
            success := reply.data^[0] = SSH_MSG_USERAUTH_SUCCESS;
            DiscardBS (reply);
        END (*IF*);

        RETURN success;

    END TryLogin;

(********************************************************************************)
(*                              DO AUTHENTICATION                               *)
(********************************************************************************)

PROCEDURE DoLogin (port: IOPort;  replyQ: PacketQueues.PacketQ;
                                            username: ARRAY OF CHAR): BOOLEAN;

    (* Requests the AUTH service, and logs a user in.   *)
    (* Returns FALSE for failure.                       *)

    VAR packet: ByteStr;  success, connectionlost: BOOLEAN;
        tries: CARDINAL;
        password: ARRAY [0..127] OF CHAR;

    BEGIN
        success := Services.RequestService (port, Services.ssh_userauth);
        IF success THEN
            packet := PacketQueues.GetPacket (replyQ);
            success := packet.data^[0] = SSH_MSG_SERVICE_ACCEPT;
            DiscardBS (packet);
            FWriteString (1, "Username: ");
            IF username[0] = Nul THEN
                ReadLine (0, username);
            ELSE
                FWriteString (1, username);  FWriteLn (1);
            END (*IF*);
        END (*IF*);

        (* First try to login without a password. *)

        connectionlost := NOT success;
        success := success AND TryLogin (port, replyQ, username, "", "none", connectionlost);
        IF connectionlost THEN
            FWriteString (1, "Connection lost");  FWriteLn (1);
            RETURN FALSE;
        END (*IF*);

        (* If that fails, try with a password.  *)

        tries := 0;
        WHILE (NOT success) AND (tries < 3) DO
            INC (tries);
            FWriteString (1, "Password: ");
            ReadKbdLine (TRUE, password);
            success := TryLogin (port, replyQ, username, password, "password", connectionlost);
            IF connectionlost THEN
                FWriteString (1, "Connection lost");  FWriteLn (1);
                RETURN FALSE;
            END (*IF*);
        END (*IF*);

        RETURN success;

    END DoLogin;

(********************************************************************************)

END ClientLogin.

