IMPLEMENTATION MODULE DiffieHellman;

        (********************************************************)
        (*                                                      *)
        (*           Diffie-Hellman exchange operation          *)
        (*                                                      *)
        (* DH algorithms not mentioned in the original standard *)
        (*  are from draft-ietf-curdle-ssh-modp-dh-sha2-09      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            12 May 2021                     *)
        (*  Last edited:        13 January 2022                 *)
        (*  Status:             working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT HashAlgs, BigNum, PublicKeys, PacketIO, Strings(*, Debug*);

FROM SYSTEM IMPORT ADR, CARD8;

FROM KeyExchange IMPORT
    (* type *)  KexType;

FROM HashAlgs IMPORT
    (* type *)  HashAlgorithm;

FROM BigNum IMPORT
    (* type *)  BN;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  SendPacket, PeerAddr;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransactionL, LogTransaction;

FROM PublicKeys IMPORT
    (* type *)  PKAType;

FROM VarStrings IMPORT
    (* type *)  ByteStr, ByteStringPtr,
    (* proc *)  BNtoByteStr, MakeBS, DiscardBS, AddByteString;

FROM Consts IMPORT
    (* const*)  SSH_MSG_KEXDH_INIT, SSH_MSG_KEXDH_REPLY;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM LowLevel IMPORT
    (* proc *)  Copy, AddOffset, EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE;

FROM STextIO IMPORT                 (* while debugging *)
    WriteChar, WriteLn;

(************************************************************************)

VAR generator, Oakley2, Oakley14, Group15: BN;

(************************************************************************)
(*               DIFFIE-HELLMAN KEY EXCHANGE - CLIENT SIDE              *)
(************************************************************************)

PROCEDURE ClientPhase1 (sess: IOPort;  VAR (*IN*) prime: BN): BN;

    (* Phase 1 of the key exchange.  We generate an ephemeral keypair   *)
    (* (private and associated public key) and send the server the      *)
    (* public key in a SSH_MSG_KEXDH_INIT message.  This keypair is     *)
    (* EPHEMERAL: it will only be used during the key exchange and      *)
    (* disposed of afterwards.                                          *)

    VAR a, e: BN;
        packet: ByteStr;
        p: ByteStringPtr;
        length: CARDINAL;

    BEGIN
        (* Client chooses random number a and sends e = g^a mod p,      *)
        (* where g and p are the Oakley parameters.  a becomes the      *)
        (* client's private key, and e the public key.                  *)

        a := BigNum.Rand(28);
        e := BigNum.ModularPower (generator, a, prime);

        (* Send e to the server in a SSH_MSG_KEXDH_INIT message.    *)

        length := BigNum.Nbytes (e);
        MakeBS (packet, length + 5);
        packet.data^[0] := SSH_MSG_KEXDH_INIT;
        length := Swap4 (length);
        Copy (ADR(length), ADR(packet.data^[1]), 4);
        p := ADR(packet.data^[5]);
        EVAL (BigNum.BNtoBytes (e, p^));
        IF SendPacket (sess, packet) THEN

            (* Save an encoding of e for later use.   *)

            PacketIO.SetString (sess, 6, BNtoByteStr(e), FALSE);
        ELSE
            BigNum.Discard (a);
            a := BigNum.Zero();
        END (*IF*);
        BigNum.Discard (e);

        (* Return our private ephemeral key, for use in phase 2. *)

        RETURN a;

    END ClientPhase1;

(************************************************************************)

(*
PROCEDURE WriteHex1 (A: CARD8);

    (* For debugging: Writes A as one-digit hex. *)

    BEGIN
        IF A > 9 THEN WriteChar (CHR(A-10+ORD('a')));
        ELSE WriteChar (CHR(A+ORD('0')));
        END (*IF*);
    END WriteHex1;

(************************************************************************)

PROCEDURE WriteHex2 (A: CARD8);

    (* For debugging: Writes A in hex. *)

    BEGIN
        WriteHex1 (A DIV 16);
        WriteHex1 (A MOD 16);
    END WriteHex2;

(************************************************************************)

PROCEDURE WriteBytes (N: CARDINAL;  VAR (*IN*) A: ARRAY OF CARD8);

    (* For debugging: Writes A in hex. *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteHex2 (A[j]);
        END (*FOR*);
    END WriteBytes;
*)

(************************************************************************)

PROCEDURE ClientPhase2 (sess: IOPort;  logID: TransactionLogID;
                            packet: ByteStr;
                            VAR (*IN*) kxprivate, prime: BN;
                            pka: PKAType;  hashalg: HashAlgorithm): HashAlgs.KeygenCtx;

    (* Phase 2 of the key exchange.  The packet contains three items:   *)
    (* the server's public key, the f value of the DH calculation, and  *)
    (* a signature.  Our job is to repeat the K and H calculation that  *)
    (* the server has already done, and verify that the signature is    *)
    (* correct.  We also want to check, using the public key, that the  *)
    (* server is known to us.                                           *)

    VAR size, j, pos: CARDINAL;
        K_S, H, part, sess_id, signature: ByteStr;
        f, K: BN;
        ctx: HashAlgs.HashCtx;
        success: BOOLEAN;

    BEGIN
        success := TRUE;

        (* The packet contains three items: the server's public key     *)
        (* K_S, the DH f value, and a signature.                        *)

        size := 0;
        Copy (ADR(packet.data^[1]), ADR(size), 4);
        size := Swap4 (size);
        MakeBS (K_S, size);
        Copy (ADR(packet.data^[5]), K_S.data, size);

        (* Do we accept the server's public key? *)

        IF NOT PublicKeys.CheckPublicKey (pka, PeerAddr(sess), K_S, logID) THEN
            DiscardBS (K_S);
            RETURN NIL;
        END (*IF*);

        pos := 5 + size;
        Copy (ADR(packet.data^[pos]), ADR(size), 4);
        INC (pos, 4);
        size := Swap4 (size);
        f := BigNum.BinToBN (packet.data^, pos, size);
        INC (pos, size);

        (* Now we can calculate the shared secret K.  *)

        K := BigNum.ModularPower (f, kxprivate, prime);

        (*BigNum.WriteBignum (K);*)         (* for comparison with server value *)

        (* The exchange hash is calculated from eight items.  Most of   *)
        (* these are already stored as PacketIO data. Put the remaining *)
        (* three in the same place.                                     *)

        PacketIO.SetString (sess, 5, K_S, TRUE);
        PacketIO.SetString (sess, 7, BNtoByteStr(f), FALSE);
        PacketIO.SetString (sess, 8, BNtoByteStr(K), FALSE);

        (* The calculation of exchange hash should ultimately be code   *)
        (* common to server and client, but for now let's do it         *)
        (* separately so that we can cross-check between the two.       *)

        MakeBS (H, HashAlgs.HashLength(hashalg));

        ctx := HashAlgs.HashInit (hashalg);
        FOR j := 1 TO 8 DO
            part := PacketIO.GetString (sess, j, TRUE);
            (*
            IF j < 4 THEN
                WriteBytes (part.size, part.data^);   WriteLn;
            END (*IF*);
            *)
            size := Swap4(part.size);
            HashAlgs.HashUpdate (ctx, size, 4);
            HashAlgs.HashUpdate (ctx, part.data^, part.size);
            DiscardBS (part);
        END (*FOR*);
        HashAlgs.HashFinal (ctx, H.data^);

        (*LogTransactionL (logID, "Have generated H");*)
        (*WriteBytes (Hlength, H.data^);   WriteLn;*)

        (* On the first key exchange, the exchange hash also becomes    *)
        (* the session id.  The following call will change the session  *)
        (* id to H only if a session id has not already been set.       *)

        sess_id := PacketIO.SetSessionID (sess, H);

        (* Now we have to check whether the signature is valid.  *)

        Copy (ADR(packet.data^[pos]), ADR(size), 4);
        INC (pos, 4);
        size := Swap4 (size);
        MakeBS (signature, size);
        Copy (ADR(packet.data^[pos]), signature.data, size);
        IF NOT PublicKeys.SignatureValid (pka, H, K_S, signature, logID) THEN
            LogTransactionL (logID, "Signature mismatch");
            success := FALSE;
        ELSE
            (* Redundant message, remove it when software more mature. *)
            LogTransactionL (logID, "Signature accepted");
        END (*IF*);

        (* Discard items we no longer need.   *)

        DiscardBS (signature);
        BigNum.Discard (f);
        DiscardBS (K_S);

        (* We assume that K and H and sess_id will be later *)
        (* discarded after the key generation calculation.  *)

        IF success THEN
            RETURN HashAlgs.InitKeyGeneration (hashalg, K, H, sess_id);
        ELSE
            BigNum.Discard (K);
            DiscardBS (H);
            DiscardBS (sess_id);
            RETURN NIL;
        END (*IF*);

    END ClientPhase2;

(************************************************************************)
(*               DIFFIE-HELLMAN KEY EXCHANGE - SERVER SIDE              *)
(************************************************************************)

PROCEDURE DiffieHellman (sess: IOPort;  packet: ByteStr;  pka: PKAType;
                           VAR (*IN*) prime: BN;
                             hashalg: HashAlgorithm): HashAlgs.KeygenCtx;

    (* The main function of this key exchange is to calculate a shared  *)
    (* key K and an exchange hash H which are known to both the server  *)
    (* and the client.  K and H are then used to calculate the keys and *)
    (* initialisation vectors for further communication.                *)

    (* Note that K itself will be thrown away once the key exchange is  *)
    (* complete.  For security reasons, this is a use-once key.         *)

    VAR K_S, signature, e, f, packet2, part, sess_id: ByteStr;
        ctx: HashAlgs.HashCtx;
        j, size: CARDINAL;
        kserver_private, kserver_public, kclient_public, K: BN;
        H: ByteStr;
        kgctx: HashAlgs.KeygenCtx;
        (*message: ARRAY [0..127] OF CHAR;*)

    BEGIN
        size := 0;              (* to avoid a compiler warning *)

        (* The packet contains an epheremal public key from the client.  *)

        Copy (ADR(packet.data^[1]), ADR(size), 4);
        size := Swap4 (size);

        (* Let e be the ByteStr form of this number. *)

        MakeBS (e, size);
        Copy (ADR(packet.data^[5]), e.data, size);
        kclient_public := BigNum.BinToBN (e.data^, 0, size);

        (* Generate our own ephemeral keypair.  Let f be the ByteStr    *)
        (* version of kserver_public.  Note: the argument to Rand       *)
        (* should really be larger, but then there's a speed problem.   *)

        kserver_private := BigNum.Rand(5);
        kserver_public := BigNum.ModularPower (generator, kserver_private, prime);
        f := BNtoByteStr (kserver_public);

        (* K is the shared key, and K_S is our host public key.  Do not *)
        (* confuse K_S with kserver_public, which is only a temporary   *)
        (* key used to calculate K, and K itself is only temporary.     *)

        K := BigNum.ModularPower (kclient_public, kserver_private, prime);
        K_S := PublicKeys.GetFormattedPublicKey(pka);

        (*BigNum.WriteBignum (K);*)         (* for comparison with client value *)

        (* The following calculation will create a hash from eight      *)
        (* items.  Four of them are already stored as PacketIO data.    *)
        (* For ease of reading the code, put the other four in the same *)
        (* place.                                                       *)

        PacketIO.SetString (sess, 5, K_S, TRUE);
        PacketIO.SetString (sess, 6, e, TRUE);
        PacketIO.SetString (sess, 7, f, TRUE);
        PacketIO.SetString (sess, 8, BNtoByteStr(K), FALSE);

        (* Next, the server generates something referred to as the      *)
        (* exchange hash H and signs it generating HS.                  *)

        (* Calculate the exchange hash H, by hashing the concatenation  *)
        (* of eight things.                                             *)

        MakeBS (H, HashAlgs.HashLength(hashalg));

        ctx := HashAlgs.HashInit (hashalg);
        FOR j := 1 TO 8 DO
            part := PacketIO.GetString (sess, j, TRUE);
            (*
            IF j < 4 THEN
                WriteBytes (part.size, part.data^);   WriteLn;
            END (*IF*);
            *)
            size := Swap4(part.size);
            HashAlgs.HashUpdate (ctx, size, 4);
            HashAlgs.HashUpdate (ctx, part.data^, part.size);
            DiscardBS (part);
        END (*FOR*);
        HashAlgs.HashFinal (ctx, H.data^);

        (*WriteBytes (Hlength, H.data^);  WriteLn;*)

        (* On the first key exchange, the exchange hash also becomes    *)
        (* the session id.  The following call will change the session  *)
        (* id to H only if a session id has not already been set.       *)

        sess_id := PacketIO.SetSessionID (sess, H);

        (* Create a signature from H. *)

        signature := PublicKeys.Sign (pka, H);

        (* Now our reply to the client will consist of the reply code,  *)
        (* then K_S, then kserver_public, then the signature.           *)

        size := 1 + 4 + K_S.size + 4 + f.size + 4 + signature.size;
        MakeBS (packet2, size);

        packet2.data^[0] := SSH_MSG_KEXDH_REPLY;
        j := 1;

        (* Fill packet2 with our public key K_S, our ephemeral public   *)
        (* key f, and the signature.                                    *)

        AddByteString (packet2, j, K_S.data^, K_S.size);
        AddByteString (packet2, j, f.data^, f.size);
        AddByteString (packet2, j, signature.data^, signature.size);

        (* Discard the variables we no longer need.  *)

        DiscardBS (K_S);
        DiscardBS (signature);
        DiscardBS (e);
        DiscardBS (f);
        BigNum.Discard (kserver_private);
        BigNum.Discard (kserver_public);
        BigNum.Discard (kclient_public);

        (*
        Debug.DebugMes ("SSH_MSG_KEXDH_REPLY packet being sent to client");
        Debug.DumpByteString (packet2);
        *)

        (* Finally we have constructed the packet.  Send it. *)

        EVAL (SendPacket (sess, packet2));

        (* We are not yet done.  We still have to generate six keys.    *)
        (* For now I'm choosing to initialise that operation by         *)
        (* telling the key generation software the values of K and H    *)
        (* and sess_id, leaving the caller to finish the job.           *)

        kgctx := HashAlgs.InitKeyGeneration (hashalg, K, H, sess_id);
        RETURN kgctx;

    END DiffieHellman;

(************************************************************************)
(*                    SERVER KEY EXCHANGE OPERATION                     *)
(************************************************************************)

PROCEDURE DHSrvDoKeyExchange (sess: IOPort; kex: KexType;  pka: PKAType;
                                     packet: ByteStr): HashAlgs.KeygenCtx;

    (* Responds to a SSH_MSG_KEX_ECDH_INIT message.    *)

    VAR kgctx: HashAlgs.KeygenCtx;

    BEGIN
        kgctx := NIL;
        CASE kex OF
            |   dh1:    kgctx := DiffieHellman (sess, packet, pka, Oakley2, sha1);
            |   dh14:   kgctx := DiffieHellman (sess, packet, pka, Oakley14, sha1);
            |   dh14_sha256:
                        kgctx := DiffieHellman (sess, packet, pka, Oakley14, sha2_256);
            |   dh15_sha512:
                        kgctx := DiffieHellman (sess, packet, pka, Group15, sha2_512);
        ELSE
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
        RETURN kgctx;
    END DHSrvDoKeyExchange;

(************************************************************************)
(*                    CLIENT KEY EXCHANGE OPERATION                     *)
(************************************************************************)

PROCEDURE DHClientStartKeyExchange (sess: IOPort; kex: KexType;
                                  pka: PKAType;  ID: TransactionLogID): BN;

    (* Client phase 1 of the key exchange.  *)

    BEGIN
        CASE kex OF
            |   dh1:    RETURN ClientPhase1 (sess, Oakley2);
            |   dh14:   RETURN ClientPhase1 (sess, Oakley14);
            |   dh14_sha256:
                        RETURN ClientPhase1 (sess, Oakley14);
            |   dh15_sha512:
                        RETURN ClientPhase1 (sess, Group15);
        ELSE
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
    END DHClientStartKeyExchange;

(************************************************************************)

PROCEDURE DHClientKxphase2 (sess: IOPort;  ID: TransactionLogID;
                           packet: ByteStr;  kex: KexType;  pka: PKAType;
                             VAR (*IN*) kxprivate: BN): HashAlgs.KeygenCtx;

    (* Client phase 2 of the key exchange.  The kxprivate parameter is  *)
    (* the number returned by phase 1.                                  *)

    BEGIN
        CASE kex OF
            |   dh1:    RETURN ClientPhase2 (sess, ID, packet, kxprivate,
                                                     Oakley2, pka, sha1);
            |   dh14:   RETURN ClientPhase2 (sess, ID, packet, kxprivate,
                                                    Oakley14, pka, sha1);
            |   dh14_sha256:
                        RETURN ClientPhase2 (sess, ID, packet, kxprivate,
                                                    Oakley14, pka, sha2_256);
            |   dh15_sha512:
                        RETURN ClientPhase2 (sess, ID, packet, kxprivate,
                                                    Group15, pka, sha2_512);
        ELSE
            RETURN NIL;
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
    END DHClientKxphase2;

(************************************************************************)
(*                         MODULE INITIALISATION                        *)
(************************************************************************)

BEGIN
    (* These values are from RFC 2409 and RFC 3526.  *)

    generator := BigNum.MakeBignum (2);

    Oakley2 := BigNum.HexToBN (
                   "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
                 + "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
                 + "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
                 + "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
                 + "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE65381"
                 + "FFFFFFFF FFFFFFFF");

    Oakley14 := BigNum.HexToBN (
                  "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
                + "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
                + "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
                + "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
                + "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D"
                + "C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F"
                + "83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D"
                + "670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B"
                + "E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9"
                + "DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510"
                + "15728E5A 8AACAA68 FFFFFFFF FFFFFFFF");

    Group15 := BigNum.HexToBN (
                  "FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1"
                + "29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD"
                + "EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245"
                + "E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED"
                + "EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D"
                + "C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F"
                + "83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D"
                + "670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B"
                + "E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9"
                + "DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510"
                + "15728E5A 8AAAC42D AD33170D 04507A33 A85521AB DF1CBA64"
                + "ECFB8504 58DBEF0A 8AEA7157 5D060C7D B3970F85 A6E1E4C7"
                + "ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226 1AD2EE6B"
                + "F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C"
                + "BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31"
                + "43DB5BFC E0FD108E 4B82D120 A93AD2CA FFFFFFFF FFFFFFFF");

FINALLY
    BigNum.Discard (Group15);
    BigNum.Discard (Oakley14);
    BigNum.Discard (Oakley2);
END DiffieHellman.

