(**************************************************************************)
(*                                                                        *)
(*  SFTP client                                                           *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FDFiles;

        (********************************************************)
        (*                                                      *)
        (*                    File operations                   *)
        (*               This is the client version             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            1 December 2021                 *)
        (*  Last edited:        21 September 2022               *)
        (*  Status:             Looks complete and working      *)
        (*                                                      *)
        (********************************************************)



IMPORT Strings, FileOps, OS2;

FROM SYSTEM IMPORT CARD8, LOC, CAST, ADR, MOVE;

FROM Channels IMPORT
    (* proc *)  SendData, ReceivePacket;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, AddCard, AddByte, AddByteString, AddCharString,
                GetCard, GetByteString, GetCharString;

FROM Arith64 IMPORT
    (* type *)  CARD64,
    (* proc *)  Add64;

FROM CodePage IMPORT
    (* proc *)  TranslateToUTF8r, TranslateFromUTF8r;

FROM MiscFuncs IMPORT
    (* proc *)  EVAL;

FROM FileOps IMPORT
    (* type *)  FilenameString, ChanId, DirectoryEntry, FileAttribute,
    (* proc *)  FirstDirEntry, DirSearchDone;

FROM Consts IMPORT
    (* const*)  SSH_FXP_OPEN, SSH_FXP_CLOSE, SSH_FXP_READ, SSH_FXP_WRITE,
                SSH_FXP_OPENDIR, SSH_FXP_READDIR, SSH_FXP_REMOVE,
                SSH_FXP_MKDIR, SSH_FXP_RMDIR, SSH_FXP_RENAME, SSH_FXP_STATUS,
                SSH_FXP_HANDLE, SSH_FXP_DATA, SSH_FXP_NAME;

(************************************************************************)

CONST
    Nul = CHR(0);
    STD_IN = 0;
    STD_OUT = 1;

(************************************************************************)
(*                   MISCELLANEOUS USEFUL PROCEDURES                    *)
(************************************************************************)

PROCEDURE WriteString (str: ARRAY OF CHAR);

    (* Write character string to standard output. *)

    BEGIN
        FileOps.FWriteString (STD_OUT, str);
    END WriteString;

(************************************************************************)

PROCEDURE WriteLn;

    (* Write line terminator to standard output. *)

    BEGIN
        FileOps.FWriteLn (STD_OUT);
    END WriteLn;

(************************************************************************)

PROCEDURE WriteChar (ch: CHAR);

    BEGIN
        FileOps.FWriteChar (STD_OUT, ch);
    END WriteChar;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal to standard output. *)

    BEGIN
        FileOps.FWriteLJCard (STD_OUT, N);
    END WriteCard;

(************************************************************************)

PROCEDURE IsWild (VAR (*IN*) filespec: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff filespec contains '*' or '?' characters.  *)

    VAR pos: CARDINAL;  found: BOOLEAN;

    BEGIN
        Strings.FindNext ('*', filespec, 0, found, pos);
        IF NOT found THEN
            Strings.FindNext ('?', filespec, 0, found, pos);
        END (*IF*);
        RETURN found;
    END IsWild;

(************************************************************************)

PROCEDURE MakeMask (VAR (*IN*) dir, param: ARRAY OF CHAR;
                                    VAR (*OUT*) mask: ARRAY OF CHAR);

    (* Makes a directory search from dir\param. *)

    VAR D: DirectoryEntry;

    BEGIN
        Strings.Assign (dir, mask);
        IF NOT LocalCombinePath (mask, param) THEN
            WriteString ("Invalid parameter");  WriteLn;
            RETURN;
        END (*IF*);
        IF NOT IsWild (mask) THEN
            (* What we do depends on whether mask specifies a directory. *)
            IF FirstDirEntry (mask, TRUE, FALSE, FALSE, D)
                                    AND (directory IN D.attr) THEN
                Strings.Append ("\*", mask);
            END (*IF*);
            DirSearchDone (D);
        END (*IF*);
    END MakeMask;

(************************************************************************)
(*                      SIMPLIFYING A PATH STRING                       *)
(************************************************************************)

PROCEDURE StripTail (VAR (*INOUT*) name: ARRAY OF CHAR): BOOLEAN;

    (* Deletes the last subdirectory specification from name, e.g.      *)
    (* we change A\B\C\D to A\B\C.  If name happens to end with a '\'   *)
    (* then we just remove the '\'.  If there is no '\' anywhere in the *)
    (* name we can't go up in the tree, so we report an error.  If the  *)
    (* found '\' is the first character of name, this too is a special  *)
    (* case.                                                            *)

    VAR pos: CARDINAL;  found: BOOLEAN;

    BEGIN
        IF name[0] <> Nul THEN
            Strings.FindPrev ('\', name, Strings.Length(name)-1, found, pos);
            IF found THEN
                IF pos = 0 THEN
                    IF name[1] = Nul THEN
                        RETURN FALSE;
                    END (*IF*);
                ELSE
                    name[pos] := Nul;
                END (*IF*);
            ELSE
                RETURN FALSE;
            END (*IF*);
        END (*IF*);
        RETURN TRUE;
    END StripTail;

(************************************************************************)

PROCEDURE LocalCombinePath (VAR (*INOUT*) path: ARRAY OF CHAR;
                                            rest: ARRAY OF CHAR): BOOLEAN;

    (* Changes path to the equivalant of path/rest, or just to rest if  *)
    (* rest is an absolute path string.  We assume that path is         *)
    (* syntactically acceptable already, but rest might contain '..'    *)
    (* components.  Returns FALSE if syntax is bad, but we don't        *)
    (* guarantee that the result is an existing file or directory.      *)

    VAR N, pos: CARDINAL;
        head: FilenameString;
        found: BOOLEAN;

    BEGIN
        (* Initial pass: change every '/' in rest to '\'. *)

        pos := 0;
        REPEAT
            Strings.FindNext ('/', rest, pos, found, pos);
            IF found THEN
                rest[pos] := '\';
            END (*IF*);
        UNTIL NOT found;

        (* Remove final '\' from rest, if present.  *)

        N := Strings.Length(rest);
        IF N > 0 THEN
            DEC (N);
            IF rest[N] = '\' THEN
                rest[N] := Nul;
            END (*IF*);
        END (*IF*);

        (* The original path always starts with X:\ where X is a drive  *)
        (* letter.  The other parameter might also have this form.      *)

        IF rest[1] = ':' THEN
            IF rest[2] <> '\' THEN
                (* Let's be generous and assume that a \ was intended.  *)
                Strings.Insert ('\', 2, rest);
            END (*IF*);
            Strings.Assign (rest, path);
            RETURN TRUE;
        END (*IF*);

        (* Next possibility: an absolute path with no drive letter. *)

        IF rest[0] = '\' THEN
            path[2] := Nul;
            Strings.Append (rest, path);
            RETURN TRUE;
        END (*IF*);

        (* Now we know that rest does not start or end with '\'.    *)

        WHILE rest[0] <> Nul DO
            Strings.FindNext ('\', rest, 0, found, pos);
            Strings.Assign (rest, head);
            IF found THEN
                head[pos] := Nul;
                Strings.Delete (rest, 0, pos+1);
            ELSE
                rest[0] := Nul;
            END (*IF*);

            IF Strings.Equal (head, '..') THEN
                IF NOT StripTail (path) THEN
                    RETURN FALSE;
                END (*IF*);
            ELSE
                Strings.Append ('\', path);
                Strings.Append (head, path);
            END (*IF*);
        END (*WHILE*);

        RETURN TRUE;

    END LocalCombinePath;

(************************************************************************)

PROCEDURE RemoteCombinePath (VAR (*INOUT*) path: ARRAY OF CHAR;
                                                    rest: ARRAY OF CHAR);

    (* Changes path to the equivalant of path/rest, or just to rest if  *)
    (* rest is an absolute path string.  This is very different         *)
    (* from the local case, because most of the work will be done when  *)
    (* we subsequently call FTPCommands.ReducePath.                     *)

    BEGIN
        IF rest[0] = '/' THEN
            (* An absolute path, so old path is irrelevant. *)
            Strings.Assign (rest, path);
        ELSIF rest[0] <> Nul THEN
            IF path[LENGTH(path)-1] <> '/' THEN
                Strings.Append ('/', path);
            END (*IF*);
            Strings.Append (rest, path);
        END (*IF*);
    END RemoteCombinePath;

(************************************************************************)
(*                    MESSAGES TO AND FROM SERVER                       *)
(************************************************************************)

PROCEDURE SimpleSend (channel: CARDINAL;  cmd: CARD8;
                             VAR (*INOUT*) id: CARDINAL;
                             VAR (*IN*) data: ARRAY OF LOC;  N: CARDINAL);

    (* Many SFTP messages consist of a one-byte command, an id, and     *)
    (* one item of variable-length data.  We handle that special case   *)
    (* in this procedure.                                               *)

    VAR L, pos: CARDINAL;
        pkt: ByteStr;

    BEGIN
        L := 13 + N;
        MakeBS (pkt, L+4);
        pos := 0;
        AddCard (pkt, pos, L);
        AddByte (pkt, pos, cmd);
        AddCard (pkt, pos, id);
        AddByteString (pkt, pos, data, N);
        EVAL (SendData (channel, pkt));
        <* COVERFLOW - *>
        INC (id);               (* allow it to wrap *)
        <* COVERFLOW + *>
    END SimpleSend;

(************************************************************************)

PROCEDURE GetStatusReply (channel: CARDINAL): BOOLEAN;

    (* Gets a STATUS packet, and writes error message if needed.    *)
    (* Returns FALSE on failure.                                    *)

    VAR reply: ByteStr;
        pos, status: CARDINAL;
        message: ARRAY [0..127] OF CHAR;
        success: BOOLEAN;
        msg: CARD8;

    BEGIN
        success := ReceivePacket (channel, reply);
        IF NOT success THEN
            RETURN FALSE;
        END (*IF*);
        pos := 0;
        EVAL (GetCard (reply, pos));
        msg := reply.data^[pos];  INC(pos);
        IF msg = SSH_FXP_STATUS THEN
            EVAL (GetCard (reply, pos));    (* id *)
            status := GetCard (reply, pos);
            IF status > 1 THEN
                GetCharString (reply, pos, message);
                WriteString (message);  WriteLn;
            END (*IF*);
        ELSE
            WriteString ("Unexpected packet from server");
            WriteLn;
            success := FALSE;
        END (*IF*);
        DiscardBS (reply);
        RETURN success;
    END GetStatusReply;

(************************************************************************)
(*                          FILE ATTRIBUTES                             *)
(************************************************************************)

PROCEDURE SkipAttrs (pkt: ByteStr;  VAR (*INOUT*) pos: CARDINAL);

    (* Skips over an ATTRS response. *)

    VAR count, flags: CARDINAL;
        string: ARRAY [0..255] OF CHAR;

    BEGIN
        flags := GetCard (pkt, pos);
        IF ODD(flags) THEN
            (* 64-bit size *)
            EVAL (GetCard(pkt, pos));
            EVAL (GetCard(pkt, pos));
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            (* UID and GID *)
            EVAL (GetCard(pkt, pos));
            EVAL (GetCard(pkt, pos));
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            (* permissions *)
            EVAL (GetCard(pkt, pos));
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            (* two time fields *)
            EVAL (GetCard(pkt, pos));
            EVAL (GetCard(pkt, pos));
        END (*IF*);
        flags := flags DIV 2;
        count := 0;
        IF ODD(flags) THEN
            (* extended count *)
            count := GetCard(pkt, pos);
        END (*IF*);
        WHILE count > 0 DO
            GetCharString (pkt, pos, string);   (* type *)
            GetCharString (pkt, pos, string);   (* data *)
            DEC (count);
        END (*WHILE*);
    END SkipAttrs;

(************************************************************************)
(*                      SENDING AND RECEIVING FILES                     *)
(************************************************************************)

PROCEDURE GetHandle (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                        filename: ARRAY OF CHAR;  write: BOOLEAN;
                        VAR (*OUT*) handle: FileHandle): CARDINAL;

    (* Gets a read or write handle from the server.  The function   *)
    (* result is the handle length in bytes, or 0 if the operation  *)
    (* fails.  Assumption: filename is in our local charset.        *)

    VAR pos, N, pflags, handlelength: CARDINAL;
        pkt, reply: ByteStr;
        errmes: ARRAY [0..79] OF CHAR;
        msg: CARD8;

    BEGIN
        TranslateToUTF8r (filename);
        (*
        WriteString ("GetHandle (");  WriteString (filename);
        WriteString (")");  WriteLn;
        *)
        IF write THEN pflags := 01AH ELSE pflags := 1 END(*IF*);

        (* That 01AH means write+creat+trunc, meaning create if it  *)
        (* doesn't exist, truncate if it does.                      *)

        N := 1 + 4 + 4 + LENGTH(filename) + 4 + 4;
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_OPEN);
        AddCard (pkt, pos, id);
        AddCharString (pkt, pos, filename);
        AddCard (pkt, pos, pflags);
        AddCard (pkt, pos, 0);          (* dummy ATTRS subrecord *)
        EVAL (SendData (channel, pkt));
        <* COVERFLOW - *>
        INC (id);               (* allow it to wrap *)
        <* COVERFLOW + *>

        (* A handle should come back in the reply. *)

        IF (NOT ReceivePacket (channel, reply)) OR (reply.size = 0) THEN
            msg := 0;
        ELSE
            pos := 0;
            EVAL (GetCard (reply, pos));
            msg := reply.data^[pos];  INC(pos);
        END (*IF*);

        IF msg = 0 THEN
            WriteString ("Connection lost");  WriteLn;
            handlelength := 0;
        ELSIF msg = SSH_FXP_HANDLE THEN
            EVAL (GetCard (reply, pos));    (* id *)
            handlelength := GetByteString (reply, pos, handle);
        ELSIF msg = SSH_FXP_STATUS THEN
            EVAL (GetCard (reply, pos));    (* id *)
            IF GetCard (reply, pos) <> 0 THEN    (* code *)
                GetCharString (reply, pos, errmes);
                WriteString (errmes);  WriteLn;
            END (*IF*);
            handlelength := 0;
        ELSE
            WriteString ("Open failure, unknown reason");  WriteLn;
            handlelength := 0;
        END (*IF*);
        DiscardBS (reply);

        RETURN handlelength;

    END GetHandle;

(************************************************************************)

PROCEDURE CloseHandle (channel: CARDINAL;  VAR (*INUT*) id: CARDINAL;
                                        L: CARDINAL;  handle: FileHandle);

    (* Closes a file or directory handle.  L is the length of handle.  *)

    VAR reply: ByteStr;

    BEGIN
        SimpleSend (channel, SSH_FXP_CLOSE, id, handle, L);

        (* Receive the status reply to the CLOSE, but there's nothing   *)
        (* we can do regardless of the status in the reply.             *)

        IF ReceivePacket (channel, reply) THEN
            DiscardBS (reply);
        END (*IF*);

    END CloseHandle;

(************************************************************************)

PROCEDURE DownloadFile (channel: CARDINAL;  VAR (*INUT*) id: CARDINAL;
                          L: CARDINAL;  handle: FileHandle;  cid: ChanId);

    (* Copies file from handle to cid.  L is the length of handle.  *)

    CONST blocksize = 16384;

    VAR offset: CARD64;
        N, pos, datalength, actual: CARDINAL;
        pkt, reply: ByteStr;
        buffer: ARRAY [0..blocksize-1] OF CARD8;
        msg: CARD8;

    BEGIN
        offset := CARD64{0,0};
        REPEAT
            N := 17 + 4 + L;
            MakeBS (pkt, N+4);
            pos := 0;
            AddCard (pkt, pos, N);
            AddByte (pkt, pos, SSH_FXP_READ);
            AddCard (pkt, pos, id);
            AddByteString (pkt, pos, handle, L);
            AddCard (pkt, pos, offset.high);
            AddCard (pkt, pos, offset.low);
            AddCard (pkt, pos, blocksize);
            EVAL (SendData (channel, pkt));
            <* COVERFLOW - *>
            INC (id);
            <* COVERFLOW + *>

            (* The reply should be a DATA response.  Treat anything  *)
            (* else as end-of-file.                                  *)

            IF ReceivePacket (channel, reply) THEN
                pos := 0;
                EVAL (GetCard (reply, pos));
                msg := reply.data^[pos];  INC(pos);
                IF msg = SSH_FXP_DATA THEN
                    EVAL (GetCard (reply, pos));    (* id *)
                    datalength := GetByteString (reply, pos, buffer);
                    OS2.DosWrite (cid, ADR(buffer), datalength, actual);
                    Add64 (offset, datalength);
                ELSE
                    datalength := 0;
                END (*IF*);
                DiscardBS (reply);
            ELSE
                datalength := 0;
            END (*IF*);

        UNTIL datalength = 0;

    END DownloadFile;

(************************************************************************)

PROCEDURE UploadFile (channel: CARDINAL;  VAR (*INUT*) id: CARDINAL;
                          L: CARDINAL;  handle: FileHandle;  cid: ChanId);

    (* Copies file from cid to handle.  L is the length of handle.  *)

    CONST blocksize = 16384;

    VAR offset: CARD64;
        N, pos, actual, status: CARDINAL;
        pkt, reply: ByteStr;
        buffer: ARRAY [0..blocksize-1] OF CARD8;
        message: ARRAY [0..127] OF CHAR;
        msg: CARD8;
        success: BOOLEAN;

    BEGIN
        offset := CARD64{0,0};
        REPEAT
            N := OS2.DosRead (cid, ADR(buffer), blocksize, actual);
            IF N <> 0 THEN
                actual := 0;
            END (*IF*);
            IF actual > 0 THEN
                N := 1 + 4 + (4 + L) + 8 + (4 + actual);
                MakeBS (pkt, N+4);
                pos := 0;
                AddCard (pkt, pos, N);
                AddByte (pkt, pos, SSH_FXP_WRITE);
                AddCard (pkt, pos, id);
                AddByteString (pkt, pos, handle, L);
                AddCard (pkt, pos, offset.high);
                AddCard (pkt, pos, offset.low);
                AddCard (pkt, pos, actual);
                MOVE (ADR(buffer), ADR(pkt.data^[pos]), actual);
                EVAL (SendData (channel, pkt));
                <* COVERFLOW - *>
                INC (id);
                <* COVERFLOW + *>

                (* The reply should be a STATUS response. *)

                success := ReceivePacket (channel, reply);
                IF success THEN
                    pos := 0;
                    EVAL (GetCard (reply, pos));
                    msg := reply.data^[pos];  INC(pos);
                    IF msg = SSH_FXP_STATUS THEN
                        EVAL (GetCard (reply, pos));    (* id *)
                        status := GetCard (reply, pos);
                        IF status > 1 THEN
                            GetCharString (reply, pos, message);
                            WriteString (message);  WriteLn;
                        END (*IF*);
                    ELSE
                        WriteString ("Unexpected packet from server");
                        WriteLn;
                    END (*IF*);
                    DiscardBS (reply);
                    Add64 (offset, actual);
                ELSE
                    actual := 0;
                END (*IF*);
            END (*IF*);

        UNTIL actual = 0;

   END UploadFile;

(************************************************************************)
(*                               DELETION                               *)
(************************************************************************)

PROCEDURE DoDelete (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                                            filename: ARRAY OF CHAR);

    (* Deletes a file, if permissions allow it. *)
    (* Assumption: path is expressed in UTF-8.  *)

    BEGIN
        SimpleSend (channel, SSH_FXP_REMOVE, id, filename, LENGTH(filename));
        EVAL (GetStatusReply (channel));
    END DoDelete;

(************************************************************************)
(*                  CREATING OR DESTROYING A DIRECTORY                  *)
(************************************************************************)

PROCEDURE MakeDirectory (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                                            dirname: ARRAY OF CHAR);

    (* Creates a new directory.  *)

    VAR N, pos: CARDINAL;
        pkt: ByteStr;

    BEGIN
        N := 1 + 4 + 4 + LENGTH(dirname) + 4;
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_MKDIR);
        AddCard (pkt, pos, id);
        AddCharString (pkt, pos, dirname);
        AddCard (pkt, pos, 0);          (* dummy ATTRS subrecord *)
        EVAL (SendData (channel, pkt));
        <* COVERFLOW - *>
        INC (id);               (* allow it to wrap *)
        <* COVERFLOW + *>

        EVAL (GetStatusReply (channel));

    END MakeDirectory;

(************************************************************************)

PROCEDURE RemoveDirectory (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                                            dirname: ARRAY OF CHAR);

    (* Deletes a directory.  *)

    BEGIN
        SimpleSend (channel, SSH_FXP_RMDIR, id, dirname, LENGTH(dirname));
        EVAL (GetStatusReply (channel));
    END RemoveDirectory;

(************************************************************************)
(*                              RENAMING                                *)
(************************************************************************)

PROCEDURE DoRename (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                                VAR (*IN*) path1, path2: ARRAY OF CHAR);

    (* Renames path1 to path2.  The legality checks are done by the server. *)

    VAR N, pos: CARDINAL;
        pkt: ByteStr;

    BEGIN
        N := 17 + LENGTH(path1) + LENGTH(path2);
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_RENAME);
        AddCard (pkt, pos, id);
        AddCharString (pkt, pos, path1);
        AddCharString (pkt, pos, path2);
        EVAL (SendData (channel, pkt));
        <* COVERFLOW - *>
        INC (id);               (* allow it to wrap *)
        <* COVERFLOW + *>

        EVAL (GetStatusReply (channel));

    END DoRename;

(************************************************************************)
(*                          DIRECTORY LISTING                           *)
(************************************************************************)

PROCEDURE DirListing (channel: CARDINAL;  VAR (*INOUT*) id: CARDINAL;
                        longform: BOOLEAN;  path: ARRAY OF CHAR): BOOLEAN;

    (* Does a directory listing of the server end directory.    *)
    (* Assumption: path is expressed in UTF-8.                  *)
    (* Returns FALSE if communication failure.                  *)

    VAR pkt, reply: ByteStr;
        N, pos, handlelength, count: CARDINAL;
        msg: CARD8;
        success: BOOLEAN;
        filename, longname: FilenameString;
        handle: FileHandle;

    BEGIN

        (* Send an OPENDIR request to the server. *)

        SimpleSend (channel, SSH_FXP_OPENDIR, id, path, LENGTH(path));

        (* Get a handle in the reply. *)

        success := ReceivePacket (channel, reply);
        IF NOT success THEN
            RETURN FALSE;
        END (*IF*);
        pos := 0;
        EVAL (GetCard (reply, pos));
        msg := reply.data^[pos];  INC(pos);
        IF msg <> SSH_FXP_HANDLE THEN
            WriteString ("No files");  WriteLn;
            DiscardBS (reply);
            RETURN TRUE;
        END (*IF*);
        EVAL (GetCard (reply, pos));    (* id *)
        handlelength := GetByteString (reply, pos, handle);
        DiscardBS (reply);

        (* Keep sending READDIR requests until there's nothing more. *)

        LOOP
            (* SimpleSend (channel, SSH_FXP_READDIR, id, handle, handlelength; *)
            N := 13 + handlelength;
            MakeBS (pkt, N+4);
            pos := 0;
            AddCard (pkt, pos, N);
            AddByte (pkt, pos, SSH_FXP_READDIR);
            AddCard (pkt, pos, id);
            AddByteString (pkt, pos, handle, handlelength);
            EVAL (SendData (channel, pkt));
            <* COVERFLOW - *>
            INC (id);
            <* COVERFLOW + *>

            (* The reply should be a NAME response.  Treat anything     *)
            (* else as end-of-listing.                                  *)

            IF NOT ReceivePacket (channel, reply) THEN
                RETURN FALSE;
            END (*IF*);
            pos := 0;
            EVAL (GetCard (reply, pos));
            msg := reply.data^[pos];  INC(pos);
            IF msg <> SSH_FXP_NAME THEN
                DiscardBS (reply);
                EXIT (*LOOP*);
            END (*IF*);
            EVAL (GetCard (reply, pos));    (* id *)
            count := GetCard (reply, pos);
            WHILE count > 0 DO
                GetCharString (reply, pos, filename);
                GetCharString (reply, pos, longname);
                TranslateFromUTF8r (filename);
                TranslateFromUTF8r (longname);
                (* Skip over ATTRS part of response. *)
                SkipAttrs (reply, pos);

                (* At last, we have something to list. *)

                IF longform THEN
                    WriteString (longname);
                ELSE
                    WriteString (filename);     (* THIS CAN BE IMPROVED *)
                END (*IF*);
                WriteLn;
                DEC (count);
            END (*WHILE*);
            DiscardBS (reply);

        END (*LOOP*);

        (* Finally, send a CLOSE request.  *)

        CloseHandle (channel, id, handlelength, handle);
        RETURN TRUE;

    END DirListing;

(************************************************************************)

END FDFiles.

