(**************************************************************************)
(*                                                                        *)
(*  SFTP client                                                           *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FtpCommands;

        (********************************************************)
        (*                                                      *)
        (*       Command interpreter for sftp client            *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            27 November 2021                *)
        (*  Last edited:        21 September 2022               *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT FileOps, Strings, PacketQueues, OS2, INIData;

FROM SYSTEM IMPORT CARD8, ADR;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  FilenameString, DirectoryEntry, FileAttribute, ChanId,
    (* proc *)  FileExists, FirstDirEntry, NextDirEntry, DirSearchDone,
                DeleteFile, OpenNewFile, OpenOldFile, CloseFile;

FROM FDFiles IMPORT
    (* type *)  FileHandle,
    (* proc *)  WriteString, WriteCard, WriteLn, LocalCombinePath, RemoteCombinePath, DirListing,
                MakeMask, GetHandle, DownloadFile, UploadFile, CloseHandle, DoDelete,
                DoRename, MakeDirectory, RemoveDirectory;

FROM Channels IMPORT
    (* proc *)  ClntOpenAChannel, RequestSubsystem, CloseChannel,
                IsClosedChannel, SendData, ReceivePacket;

FROM Consts IMPORT
    (* const*)  SSH_FXP_REALPATH, SSH_FXP_NAME, SSH_FXP_STAT, SSH_FXP_ATTRS,
                SSH_FXP_STATUS, SSH_FXP_INIT;

FROM PacketIO IMPORT
    (* type *)  IOPort;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, AddByte, AddCard, AddCharString, AddByteString,
                GetCard, GetCharString, GetByteString;

FROM SysClock IMPORT
    (* type *)  DateTime,
    (* proc *)  GetClock;

FROM MyClock IMPORT
    (* proc *)  UnpackDateTime, FormatDateTime, FormatDateTime12hr;

FROM CodePage IMPORT
    (* proc *)  TranslateToUTF8r, TranslateFromUTF8r;

FROM Conversions IMPORT
    (* proc *)  Card64ToStringTS, CardinalToString;

FROM MiscFuncs IMPORT
    (* proc *)  EVAL, ToLower;

FROM LowLevel IMPORT
    (* proc *)  IAND;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE CharSet = SET OF CHAR;

CONST
    ourversion = 3;

    STD_IN = 0;
    STD_OUT = 1;

    Nul = CHR(0);
    Digits = CharSet{'0'..'9'};

TYPE
    (* The session record.  The fields are:                             *)
    (*      chan            channel number                              *)
    (*      id              identifier for next SFTP request            *)
    (*      user            user account                                *)
    (*      localdir        current directory at local end              *)
    (*      remotedir       current directory at remote end             *)

    Session = POINTER TO
                    RECORD
                        chan: CARDINAL;
                        id: CARDINAL;
                        connectionlost: BOOLEAN;
                        localdir,
                        remotedir: FileOps.FilenameString;
                    END (*RECORD*);

VAR
    (* Format string for formatting a timestamp. *)

    DTFormat: ARRAY [0..31] OF CHAR;

    (* Flag to say we are using a twelve-hour clock. *)

    TwelveHour: BOOLEAN;

    ThousandsSeparator: CHAR;

(************************************************************************)
(*                   MISCELLANEOUS USEFUL PROCEDURES                    *)
(************************************************************************)

PROCEDURE SendIt (S: Session;  pkt: ByteStr): BOOLEAN;

    (* Sends the packet, and also increments the id. *)

    VAR success: BOOLEAN;

    BEGIN
        success := SendData (S^.chan, pkt);
        IF success THEN
            <* COVERFLOW - *>
            INC (S^.id);
            <* COVERFLOW + *>
        END (*IF*);
        RETURN success;
    END SendIt;

(************************************************************************)

PROCEDURE ReducePath (S: Session;  VAR (*INOUT*) path: ARRAY OF CHAR): BOOLEAN;

    (* Reduces a remote filename string to canonical form, with things  *)
    (* like '.' and '..' removed.  As a useful side-effect translates   *)
    (* the result into UTF-8.                                           *)

    VAR pkt, reply: ByteStr;
        N, pos: CARDINAL;
        msg: CARD8;

    BEGIN
        TranslateToUTF8r (path);

        (* Send a REALPATH request to the server. *)

        N := 9 + Strings.Length(path);
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_REALPATH);
        AddCard (pkt, pos, S^.id);
        AddCharString (pkt, pos, path);
        IF NOT SendIt (S, pkt) THEN
            S^.connectionlost := TRUE;
            RETURN FALSE;
        END (*IF*);

        (* Get the reply. *)

        IF NOT ReceivePacket (S^.chan, reply) THEN
            S^.connectionlost := TRUE;
            RETURN FALSE;
        END (*IF*);
        pos := 0;
        EVAL (GetCard (reply, pos));
        msg := reply.data^[pos];  INC(pos);
        EVAL (GetCard (reply, pos));            (* id *)
        IF msg = SSH_FXP_NAME THEN
            EVAL (GetCard (reply, pos));        (* dummy count *)
            GetCharString (reply, pos, path);
            DiscardBS (reply);
            RETURN TRUE;
        ELSIF msg = SSH_FXP_STATUS THEN
            N := GetCard (reply, pos);        (* error code *)
            DiscardBS (reply);
            RETURN N = 0;
        ELSE
            DiscardBS (reply);
            S^.connectionlost := TRUE;
            RETURN FALSE;
        END (*IF*);

    END ReducePath;

(************************************************************************)

PROCEDURE IsDirectory (S: Session;  path: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE if path is a directory.  We assume path is in UTF-8. *)

    CONST dirflag = 4000H;

    VAR N, pos, flags, permissions: CARDINAL;
        pkt, reply: ByteStr;
        msg: CARD8;

    BEGIN
        (* Send a STAT request to the server. *)

        N := 13 + Strings.Length(path);
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_STAT);
        AddCard (pkt, pos, S^.id);
        AddCharString (pkt, pos, path);
        IF SendData (S^.chan, pkt) THEN
            INC (S^.id);
        ELSE
            DiscardBS (pkt);
            RETURN FALSE;
        END (*IF*);

        (* Get the reply. *)

        IF NOT ReceivePacket (S^.chan, reply) THEN
            RETURN FALSE;
        END (*IF*);
        pos := 0;
        EVAL (GetCard (reply, pos));            (* size *)
        msg := reply.data^[pos];  INC(pos);
        IF msg = SSH_FXP_ATTRS THEN
            EVAL (GetCard (reply, pos));        (* id *)
            flags := GetCard (reply, pos);
            IF ODD(flags) THEN
                (* skip size *)
                EVAL (GetCard (reply, pos));
                EVAL (GetCard (reply, pos));
            END (*IF*);
            flags := flags DIV 2;
            IF ODD(flags) THEN
                (* skip uid and gid *)
                EVAL (GetCard (reply, pos));
                EVAL (GetCard (reply, pos));
            END (*IF*);
            permissions := GetCard (reply, pos);
            DiscardBS (reply);
            RETURN IAND (permissions, dirflag) <> 0;
        ELSE
            (* Classify any error reply as "not a directory" *)

            DiscardBS (reply);
            RETURN FALSE;
        END (*IF*);
    END IsDirectory;

(************************************************************************)
(*                   THE INDIVIDUAL COMMAND HANDLERS                    *)
(************************************************************************)

PROCEDURE cd (S: Session;  param: ARRAY OF CHAR);

    (* Change remote directory.  *)

    VAR path: FilenameString;
        success: BOOLEAN;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, param);
        success := ReducePath (S, path);
        IF S^.connectionlost THEN
            WriteString ("Connection lost");  WriteLn;
            RETURN;
        END (*IF*);

        IF success AND NOT IsDirectory(S, path) THEN
            WriteString (path);  WriteString (" is not a directory. ");
            success := FALSE;
        END (*IF*);

        IF success THEN
            TranslateFromUTF8r (path);
            Strings.Assign (path, S^.remotedir);
        ELSE
            WriteString ("Cannot change directory");  WriteLn;
        END (*IF*);
    END cd;

(************************************************************************)

PROCEDURE delete (S: Session;  name: ARRAY OF CHAR);

    (* Delete a remote file.  *)

    VAR path: FilenameString;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, name);
        IF NOT ReducePath (S, path) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed specification ");  WriteString (path);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        DoDelete (S^.chan, S^.id, path);
    END delete;

(************************************************************************)

PROCEDURE dir (S: Session;  param: ARRAY OF CHAR);

    (* Remote directory listing, long form.  *)

    VAR path: FilenameString;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, param);
        IF NOT ReducePath (S, path) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed specification ");  WriteString (path);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        S^.connectionlost := NOT DirListing (S^.chan, S^.id, TRUE, path);
        IF S^.connectionlost THEN
            WriteString ("Connection lost");  WriteLn;
        END (*IF*);
    END dir;

(************************************************************************)

PROCEDURE get (S: Session;  name: ARRAY OF CHAR);

    (* Fetch a file from the server. *)

    VAR handle: FileHandle;
        cid: ChanId;
        L: CARDINAL;
        filename, localfile: FilenameString;
        IsDir: BOOLEAN;

    BEGIN
        Strings.Assign (S^.remotedir, filename);
        RemoteCombinePath (filename, name);
        L := GetHandle (S^.chan, S^.id, filename, FALSE, handle);
        (* L is the length of the handle. *)
        IF L = 0 THEN
            WriteString ("Failed to get a file handle from server");
            WriteLn;
            RETURN;
        END (*IF*);

        (* Work out the local file name. Delete the file if it exists,  *)
        (* because the fetched file will overwrite it.                  *)

        Strings.Assign (S^.localdir, localfile);
        IF NOT LocalCombinePath (localfile, name) THEN
            WriteString ("Invalid local file name");  WriteLn;
            RETURN;
        END (*IF*);
        IF FileExists (localfile, IsDir) THEN
            IF IsDir THEN
                WriteString ("Failed: attempt to overwrite a directory");
                WriteLn;
                RETURN;
            ELSE
                DeleteFile (localfile);
            END (*IF*);
        END (*IF*);
        cid := OpenNewFile (localfile, FALSE);
        DownloadFile (S^.chan, S^.id, L, handle, cid);
        CloseFile (cid);
        CloseHandle (S^.chan, S^.id, L, handle);

    END get;

(************************************************************************)

PROCEDURE help (S: Session;  param: ARRAY OF CHAR);

    (* User asked for help.  *)

    BEGIN
        EVAL(S);                    (* to suppress a compiler warning *)
        param[0] := param[0];       (* to suppress a compiler warning *)
        WriteString ("    ?     bye     cd      del    delete dir    exit   get   help");
        WriteLn;
        WriteString ("   lcd    ldel    ldelete ldir   lls    lpwd   lrm    ls    mkdir");
        WriteLn;
        WriteString ("   put    pwd     quit    ren    rename rm     rmdir  version");
        WriteLn;
    END help;

(************************************************************************)

PROCEDURE lcd (S: Session;  param: ARRAY OF CHAR);

    (* Change local directory.  *)

    VAR path: FilenameString;
        directory: BOOLEAN;

    BEGIN
        Strings.Assign (S^.localdir, path);
        IF NOT LocalCombinePath (path, param) THEN
            WriteString ("Invalid parameter");  WriteLn;
        ELSE
            (* We still have to confirm that we have a directory. *)

            IF FileExists (path, directory) AND directory THEN
                Strings.Assign (path, S^.localdir);
            ELSE
                WriteString (path);  WriteString (" is not a directory. ");
                WriteLn;
            END (*IF*);
        END (*IF*);
    END lcd;

(************************************************************************)

PROCEDURE ldel (S: Session;  param: ARRAY OF CHAR);

    (* Delete local file or directory.  *)

    VAR path: FilenameString;
        rc: CARDINAL;
        directory: BOOLEAN;

    BEGIN
        Strings.Assign (S^.localdir, path);
        IF NOT LocalCombinePath (path, param) THEN
            WriteString ("Invalid parameter");  WriteLn;
        ELSIF FileExists (path, directory) THEN
            IF directory THEN
                rc := OS2.DosDeleteDir (path);
                IF rc = 0 THEN
                    WriteString ("directory");
                END (*IF*);
            ELSE
                rc := OS2.DosDelete (path);
                IF rc = 0 THEN
                    WriteString ("file");
                END (*IF*);
            END (*IF*);
            rc := OS2.DosDelete (path);
            IF rc = 0 THEN
                WriteString (" deleted");  WriteLn;
            ELSE
                WriteString (" deletion failed");  WriteLn;
            END (*IF*);
        ELSE
            WriteString (path);  WriteString (" does not exist. ");
            WriteLn;
        END (*IF*);
    END ldel;

(************************************************************************)

PROCEDURE ldir (S: Session;  param: ARRAY OF CHAR);

    (* List local directory, long form.  *)

    (********************************************************************)

    PROCEDURE ashrcode (D: DirectoryEntry;  VAR (*OUT*) code: ARRAY OF CHAR);

        BEGIN
            Strings.Assign ("----", code);
            IF archive IN D.attr THEN code[0] := 'a' END (*IF*);
            IF system IN D.attr THEN code[1] := 's' END (*IF*);
            IF hidden IN D.attr THEN code[2] := 'h' END (*IF*);
            IF readonly IN D.attr THEN code[3] := 'r' END (*IF*);
        END ashrcode;

    (********************************************************************)

    CONST Linewidth = 80;

    VAR line: ARRAY [0..Linewidth-1] OF CHAR;

    (********************************************************************)

    PROCEDURE Spaces(N: CARDINAL);

        (* Adds N space characters to line. *)

        BEGIN
            WHILE N >= 8 DO
                Strings.Append ("        ", line);  DEC (N, 8);
            END (*WHILE*);
            WHILE N >= 4 DO
                Strings.Append ("    ", line);  DEC (N, 4);
            END (*WHILE*);
            WHILE N >= 2 DO
                Strings.Append ("  ", line);  DEC (N, 2);
            END (*WHILE*);
            WHILE N >= 1 DO
                Strings.Append (" ", line);  DEC (N);
            END (*WHILE*);
        END Spaces;

    (********************************************************************)

    CONST Sizwidth = 15;

    VAR mask, name: FilenameString;
        D: DirectoryEntry;
        DT: DateTime;
        tmpbuf: ARRAY [0..31] OF CHAR;

    BEGIN
        MakeMask (S^.localdir, param, mask);
        IF FirstDirEntry (mask, TRUE, FALSE, FALSE, D) THEN
            REPEAT
                DT := UnpackDateTime (D.datePkd, D.timePkd);
                IF TwelveHour THEN
                    FormatDateTime12hr (DT, DTFormat, FALSE, line);
                ELSE
                    FormatDateTime (DT, DTFormat, FALSE, line);
                END (*IF*);
                name := D.name;
                IF directory IN D.attr THEN
                    Spaces (Sizwidth - 5);
                    Strings.Append ("<DIR>", line);
                    Strings.Append ('\', name);
                ELSE
                    Card64ToStringTS (D.size, line, LENGTH(line),
                                            Sizwidth, ThousandsSeparator);
                END (*IF*);
                Spaces(1);
                CardinalToString (D.easize, tmpbuf, 6);
                Strings.Append (tmpbuf, line);
                Spaces(1);
                ashrcode (D, tmpbuf);
                Strings.Append (tmpbuf, line);
                Spaces(2);
                Strings.Append (D.name, line);
                WriteString (line);  WriteLn;
            UNTIL NOT NextDirEntry (D);
            DirSearchDone (D);
        ELSE
            WriteString ("<no files>");  WriteLn;
        END (*IF*);
        WriteLn;
    END ldir;

(************************************************************************)

PROCEDURE lls (S: Session;  param: ARRAY OF CHAR);

    (* List local directory, short form.  *)

    CONST
        Linewidth = 80;  NCols = 5;  Ewidth = Linewidth DIV NCols;

    VAR mask, name: FilenameString;
        L, M: CARDINAL;
        D: DirectoryEntry;

    BEGIN
        MakeMask (S^.localdir, param, mask);
        M := 0;
        IF FirstDirEntry (mask, TRUE, FALSE, FALSE, D) THEN
            REPEAT
                name := D.name;
                IF directory IN D.attr THEN
                    Strings.Append ('\', name);
                END (*IF*);
                L := Strings.Length (name);
                IF M + L >= Linewidth THEN
                    WriteLn;  M := 0;
                END (*IF*);
                WriteString (name);  INC (M, L);
                REPEAT
                    WriteString (" ");  INC(M);
                UNTIL (M = Linewidth-1) OR (M MOD Ewidth = 0);
            UNTIL NOT NextDirEntry (D);
            DirSearchDone (D);
            IF M > 0 THEN
                WriteLn;
            END (*IF*);
        ELSE
            WriteString ("<no files>");  WriteLn;
        END (*IF*);
        WriteLn;
    END lls;

(************************************************************************)

PROCEDURE lmkdir (S: Session;  dir: ARRAY OF CHAR);

    (* Make a new local directory. *)

    VAR path: FilenameString;
        rc: CARDINAL;  directory: BOOLEAN;

    BEGIN
        Strings.Assign (S^.localdir, path);
        IF NOT LocalCombinePath (path, dir) THEN
            WriteString ("Invalid parameter");  WriteLn;
        ELSIF FileExists (path, directory) THEN
            IF directory THEN
                WriteString ("directory already exists");
            ELSE
                WriteString ("there is a file with that name");
            END (*IF*);
            WriteLn;
            RETURN;
        END (*IF*);
        rc := OS2.DosCreateDir (path, NIL);
        IF rc <> 0 THEN
            WriteString ("operation failed");  WriteLn;
        END (*IF*);
    END lmkdir;

(************************************************************************)

PROCEDURE lpwd (S: Session;  param: ARRAY OF CHAR);

    BEGIN
        param[0] := param[0];       (* to suppress a compiler warning *)
        WriteString ("local directory is ");
        WriteString (S^.localdir);  WriteLn;
    END lpwd;

(************************************************************************)

PROCEDURE ls (S: Session;  param: ARRAY OF CHAR);

    (* Remote directory listing, short form.  *)

    VAR path: FilenameString;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, param);
        IF NOT ReducePath (S, path) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed specification ");  WriteString (path);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        S^.connectionlost := NOT DirListing (S^.chan, S^.id, FALSE, path);
    END ls;

(************************************************************************)

PROCEDURE mkdir (S: Session;  name: ARRAY OF CHAR);

    (* Create a directory. *)

    VAR path: FilenameString;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, name);
        IF NOT ReducePath (S, path) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed specification ");  WriteString (path);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        MakeDirectory (S^.chan, S^.id, path);
    END mkdir;

(************************************************************************)

PROCEDURE put (S: Session;  name: ARRAY OF CHAR);

    (* Upload a file to the server. *)

    VAR handle: FileHandle;
        cid: ChanId;
        L: CARDINAL;
        filename, localfile: FilenameString;

    BEGIN
        Strings.Assign (S^.localdir, localfile);
        IF NOT LocalCombinePath (localfile, name) THEN
            WriteString ("Invalid local file name");  WriteLn;
            RETURN;
        END (*IF*);
        cid := OpenOldFile (localfile, FALSE, FALSE);
        IF cid = NoSuchChannel THEN
            WriteString ("Can't access local file ");
            WriteString (localfile);  WriteLn;
            RETURN;
        END (*IF*);

        Strings.Assign (S^.remotedir, filename);
        RemoteCombinePath (filename, name);
        L := GetHandle (S^.chan, S^.id, filename, TRUE, handle);
        (* L is the length of the handle. *)
        IF L = 0 THEN
            WriteString ("Failed to get a file handle from server");
            WriteLn;
            RETURN;
        END (*IF*);
        UploadFile (S^.chan, S^.id, L, handle, cid);
        CloseFile (cid);
        CloseHandle (S^.chan, S^.id, L, handle);

    END put;

(************************************************************************)

PROCEDURE pwd (S: Session;  param: ARRAY OF CHAR);

    BEGIN
        param[0] := param[0];       (* to suppress a compiler warning *)
        WriteString ("remote directory is ");
        WriteString (S^.remotedir);  WriteLn;
    END pwd;

(************************************************************************)

PROCEDURE rename (S: Session;  param: ARRAY OF CHAR);

    VAR name2, path1, path2: FilenameString;
        pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        (* param should contain two names. *)

        Strings.FindNext (' ', param, 0, found, pos);
        IF NOT found THEN
            WriteString ("Usage: rename oldname newname");
            RETURN;
        END (*IF*);
        Strings.Assign (param, name2);
        Strings.Delete (name2, 0, pos+1);
        param[pos] := Nul;
        WHILE name2[0] = ' ' DO
            Strings.Delete (name2, 0, 1);
        END (*WHILE*);

        (* Create the two paths.  *)

        Strings.Assign (S^.remotedir, path1);
        RemoteCombinePath (path1, param);
        IF NOT ReducePath (S, path1) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed source specification ");  WriteString (path1);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        Strings.Assign (S^.remotedir, path2);
        RemoteCombinePath (path2, name2);
        IF NOT ReducePath (S, path2) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed destination specification ");
                WriteString (path2);  WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);

        DoRename (S^.chan, S^.id, path1, path2);

    END rename;

(************************************************************************)

PROCEDURE rmdir (S: Session;  name: ARRAY OF CHAR);

    (* Delete a directory. *)

    VAR path: FilenameString;

    BEGIN
        Strings.Assign (S^.remotedir, path);
        RemoteCombinePath (path, name);
        IF NOT ReducePath (S, path) THEN
            IF S^.connectionlost THEN
                WriteString ("Connection lost");  WriteLn;
            ELSE
                WriteString ("Malformed specification ");  WriteString (path);
                WriteLn;
            END (*IF*);
            RETURN;
        END (*IF*);
        RemoveDirectory (S^.chan, S^.id, path);
    END rmdir;

(************************************************************************)

PROCEDURE unknown (S: Session;  command: ARRAY OF CHAR);

    (* Handler for unknown commands. *)

    BEGIN
        EVAL(S);                    (* to suppress a compiler warning *)
        WriteString ("unknown command ");  WriteString (command);
        WriteLn;
    END unknown;

(************************************************************************)

PROCEDURE version (S: Session;  param: ARRAY OF CHAR);

    (* User wants to know our version number. *)

    BEGIN
        EVAL(S);                    (* to suppress a compiler warning *)
        param[0] := param[0];       (* to suppress a compiler warning *)
        WriteString ("SFTP version ");  WriteCard (ourversion);
        WriteLn;
    END version;

(************************************************************************)
(*                          COMMAND PARSER                              *)
(************************************************************************)

CONST NumCommands = 25;          (* not counting the exit commands  *)
                                 (* which are exit, quit, bye       *)
TYPE
    CmdType = [0..NumCommands-1];

    CommandProc = PROCEDURE (Session, ARRAY OF CHAR);
    ListEntry = RECORD
                    text: ARRAY [0..15] OF CHAR;
                    proc: CommandProc;
                END (*RECORD*);

    CommandList = ARRAY CmdType OF ListEntry;

CONST
    CmdList = CommandList
                {
                    {" ",      unknown},
                    {"?",      help},
                    {"cd",     cd},
                    {"del",    delete},
                    {"delete", delete},
                    {"dir",    dir},
                    {"get",    get},
                    {"help",   help},
                    {"lcd",    lcd},
                    {"ldel",   ldel},
                    {"ldelete", ldel},
                    {"ldir",   ldir},
                    {"lls",    lls},
                    {"lmkdir", lmkdir},
                    {"lpwd",   lpwd},
                    {"lrm",    ldel},
                    {"ls",     ls},
                    {"mkdir",  mkdir},
                    {"put",    put},
                    {"pwd",    pwd},
                    {"ren",    rename},
                    {"rename", rename},
                    {"rm",     delete},
                    {"rmdir",  rmdir},
                    {"version", version}
                };

(************************************************************************)

PROCEDURE HandleCommand (S: Session;  VAR (*IN*) command: ARRAY OF CHAR;
                                                VAR (*OUT*)mustexit: BOOLEAN);

    (* Parses and executes one command.   *)

    VAR first, result, last: CmdType;
        pos: CARDINAL;
        found: BOOLEAN;
        test: Strings.CompareResults;
        param: ARRAY [0..127] OF CHAR;

    BEGIN
        (* Strip leading spaces.  *)

        WHILE command[0] = ' ' DO
            Strings.Delete (command, 0, 1);
        END (*WHILE*);

        (* Separate out the parameter, if any. *)

        Strings.FindNext (' ', command, 0, found, pos);
        IF found THEN
            Strings.Assign (command, param);
            Strings.Delete (param, 0, pos+1);
            command[pos] := Nul;
        ELSE
            param := "";
        END (*IF*);

        (* Remove leading and trailing spaces from parameter. *)

        WHILE param[0] = ' ' DO
            Strings.Delete (param, 0, 1);
        END (*WHILE*);
        pos := Strings.Length (param);
        WHILE (pos > 0) AND (param[pos-1] = ' ') DO
            DEC (pos);  param[pos] := Nul;
        END (*WHILE*);

        (* Reduce command to lower case. *)

        ToLower (command);

        (* Treat "exit" and its synonyms separately. *)

        mustexit := Strings.Equal (command, "exit") OR Strings.Equal (command, "quit")
                                OR Strings.Equal (command, "bye")
                                    OR IsClosedChannel (S^.chan);
        IF mustexit THEN
            RETURN;
        END (*IF*);

        (* Go through the keyword list to find a match with the command.  *)
        (* In this version I'm using a binary search.                     *)

        first := MIN(CmdType);  last := MAX(CmdType);
        REPEAT
            result := (first + last) DIV 2;
            test := Strings.Compare (command, CmdList[result].text);
            IF test = Strings.less THEN
                IF result = MIN(CmdType) THEN
                    test := Strings.equal;    (* A 'match' on the command 'unknown' *)
                ELSE
                    last := result;  DEC(last);
                END (*IF*);
            ELSIF test = Strings.greater THEN
                first := result;
                IF first = MAX(CmdType) THEN
                    DEC (last);       (* to force loop exit *)
                ELSE
                    INC (first);
                END (*IF*);
            END (*IF*);
        UNTIL (test = Strings.equal) OR (first > last);

        IF test <> Strings.equal THEN
            result := 0;
        END (*IF*);

        (* Execute the command. *)

        IF result = 0 THEN
            CmdList[result].proc (S, command);
        ELSE
            CmdList[result].proc (S, param);
        END (*IF*);

    END HandleCommand;

(************************************************************************)
(*                      THE FETCH/EXECUTE LOOP                          *)
(************************************************************************)

PROCEDURE GetCommand (VAR(*OUT*) command: ARRAY OF CHAR);

    BEGIN
        FileOps.ReadLine (STD_IN, command);
    END GetCommand;

(************************************************************************)

PROCEDURE RunUserCommands (S: Session);

    (* Runs the command/response cycle of the SFTP client. *)

    VAR command: ARRAY [0..511] OF CHAR;
        mustexit: BOOLEAN;

    BEGIN
        REPEAT
            WriteString ("sftp> ");
            GetCommand (command);
            HandleCommand (S, command, mustexit);
        UNTIL mustexit OR S^.connectionlost;
    END RunUserCommands;

(************************************************************************)
(*                     STARTING A SESSION                               *)
(************************************************************************)

PROCEDURE StartFTP (S: Session);

    (* Sends an INIT packet, receives a VERSION reply.  At present,     *)
    (* though, this program supports only version 3.                    *)

    CONST
        ourversion = 3;
        N = 5;          (* payload size *)

    VAR pkt, reply: ByteStr;
        pos: CARDINAL;

    BEGIN
        MakeBS (pkt, N+4);
        pos := 0;
        AddCard (pkt, pos, N);
        AddByte (pkt, pos, SSH_FXP_INIT);
        AddCard (pkt, pos, ourversion);
        EVAL (SendIt (S, pkt));

        (* Get the reply, but don't do anything with it. *)

        S^.connectionlost := NOT ReceivePacket (S^.chan, reply);
        DiscardBS (reply);

    END StartFTP;

(************************************************************************)

PROCEDURE RunFTPSession (port: IOPort;  replyQ: PacketQueues.PacketQ);

    (* Opens a channel to the server, runs one session.   *)

    VAR ourchannel: CARDINAL;
        S: Session;

    BEGIN
        ourchannel := ClntOpenAChannel (port, replyQ);
        RequestSubsystem (ourchannel, "sftp");

        NEW (S);
        S^.chan := ourchannel;
        S^.id := 1;
        S^.connectionlost := FALSE;
        StartFTP (S);

        IF NOT S^.connectionlost THEN
            FileOps.GetWorkingDirectory (S^.localdir);
            S^.remotedir := ".";
            IF NOT ReducePath (S, S^.remotedir) THEN
                IF S^.connectionlost THEN
                    WriteString ("Connection lost");  WriteLn;
                ELSE
                    WriteString ("Internal error in REALPATH call");
                    WriteLn;
                END (*IF*);
            END (*IF*);

            RunUserCommands (S);
        END (*IF*);

        CloseChannel (ourchannel);
        DISPOSE (S);

    END RunFTPSession;

(************************************************************************)
(*                          MODULE INITIALISATION                       *)
(************************************************************************)

PROCEDURE GetLocaleData;

    (* Allow for local preferences. *)

    (* We pick up all the options from OS2.INI  *)

    VAR hini: INIData.HINI;
        (*Now: DateTime;*)
        ININame, app: ARRAY [0..15] OF CHAR;
        TimeFormat: ARRAY [0..8] OF CHAR;
        (*buffer: ARRAY [0..19] OF CHAR;*)
        pos: CARDINAL;
        found: BOOLEAN;
        iDate, iTime, sDate, sTime, sThou: ARRAY [0..1] OF CHAR;

    BEGIN
        DTFormat := "dd/mm/yy ";  TimeFormat := "HH:MM ";
        TwelveHour := FALSE;
        ThousandsSeparator := ',';
        iDate := '0';
        iTime := '0';
        sDate := '/';
        sTime := ':';

        (* Get some stuff from OS2.INI *)

        ININame := "OS2.INI";
        hini := INIData.OpenINIFile (ININame);
        IF INIData.INIValid(hini) THEN
            app := "PM_National";
            EVAL (INIData.INIGet (hini, app, "iDate", iDate));
            EVAL (INIData.INIGet (hini, app, "iTime", iTime));
            EVAL (INIData.INIGet (hini, app, "sDate", sDate));
            EVAL (INIData.INIGet (hini, app, "sThousand", sThou));
            ThousandsSeparator := sThou[0];
            EVAL (INIData.INIGet (hini, app, "sTime", sTime));
        END (*IF*);
        INIData.CloseINIFile (hini);

        (* Date format. *)

        CASE iDate[0] OF
          | '0':    DTFormat := "mm/dd/yy ";
          | '1':    DTFormat := "dd/mm/yy ";
          | '2':    DTFormat := "yy/mm/dd ";
          | '3':    DTFormat := "yy/dd/mm ";
        ELSE
            (* stick with the default *)
        END (*CASE*);
        pos := 0;  found := FALSE;
        REPEAT
            Strings.FindNext ('/', DTFormat, pos, found, pos);
            IF found THEN
                DTFormat[pos] := sDate[0];  INC(pos);
            END (*IF*);
        UNTIL NOT found;

        (* Time format.  I'd also like to adjust the order of its       *)
        (* components, but I haven't found out where that is defined.   *)

        CASE iTime[0] OF
          | '0':    TimeFormat := "HH:MMa";  TwelveHour := TRUE;
          | '1':    TimeFormat := "HH:MM ";  TwelveHour := FALSE;
        ELSE
            (* stick with the default *)
        END (*CASE*);

        pos := 0;
        REPEAT
            Strings.FindNext (':', TimeFormat, pos, found, pos);
            IF found THEN
                TimeFormat[pos] := sTime[0];  INC(pos);
            END (*IF*);
        UNTIL NOT found;
        Strings.Append (TimeFormat, DTFormat);

        (* Test output *)

        (*
        GetClock (Now);
        FormatDateTime12hr (Now, DTFormat, FALSE, buffer);
        WriteString (buffer);  WriteLn;
        *)

    END GetLocaleData;

(************************************************************************)

BEGIN
    GetLocaleData;
END FtpCommands.

