(**************************************************************************)
(*                                                                        *)
(*  SFTP client                                                           *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE sftp;

        (********************************************************)
        (*                                                      *)
        (*                     SFTP client                      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            24 November 2021                *)
        (*  Last edited:        22 September 2021               *)
        (*  Status:             OK except for termination       *)
        (*                                                      *)
        (********************************************************)



FROM SYSTEM IMPORT CAST;

IMPORT IOChan, TextIO, STextIO, Strings, OS2, PacketQueues;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM sshSession IMPORT
    (* proc *)  NewSession;

FROM ClientLogin IMPORT
    (* proc *)  DoLogin;

FROM FtpCommands IMPORT
    (* proc *)  RunFTPSession;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  InitWait, ClosePacketSession;

FROM NetDB IMPORT
    (* type *)  HostEntPtr, AddressPointerArrayPointer,
    (* proc *)  gethostbyname;

FROM Sockets IMPORT
    (* const*)  NotASocket,
    (* type *)  Socket, SockAddr, AddressFamily, SocketType,
    (* proc *)  sock_init, socket, connect, soclose;

FROM Inet2Misc IMPORT
    (* proc *)  Swap2;

FROM Internet IMPORT
    (* const*)  Zero8,
    (* proc *)  inet_addr;

FROM Inet2Misc IMPORT
    (* proc *)  NameIsNumeric;

FROM ProgName IMPORT
    (* proc *)  GetProgramName, WeAreServer;

FROM Exceptq IMPORT
    (* proc *)  InstallExceptq, UninstallExceptq;

FROM Timer IMPORT
    (* proc *)  Sleep;

FROM LowLevel IMPORT
    (* proc *)  EVAL;

(********************************************************************************)

CONST
    TESTING = FALSE;
    (*testInput = "127.0.0.1";*)
    testInput = "anonymous@localhost";
    Nul = CHR(0);

VAR
    (* Server port. *)

    ServerPort: CARDINAL;

(************************************************************************)
(*                     GETTING THE SERVER ADDRESS                       *)
(************************************************************************)

PROCEDURE ExtractIPAddress (str: ARRAY OF CHAR): CARDINAL;

    (* Converts address specification to IP address in network byte order. *)

    VAR result: CARDINAL;
        hep: HostEntPtr;
        p: AddressPointerArrayPointer;

    BEGIN
        IF NameIsNumeric (str) THEN
            result := inet_addr(str);
        ELSE
            (*
            STextIO.WriteString ("Looking up '");  STextIO.WriteString (str);
            STextIO.WriteString ("'");  STextIO.WriteLn;
            *)
            hep := gethostbyname (str);
            IF hep = NIL THEN p := NIL
            ELSE p := hep^.h_addr_list
            END (*IF*);
            IF (p = NIL) OR (p^[0] = NIL) THEN
                result := 0;
            ELSE
                result := p^[0]^;
            END (*IF*);
        END (*IF*);
        RETURN result;
    END ExtractIPAddress;

(************************************************************************)

PROCEDURE GetParam(VAR (*OUT*) username: ARRAY OF CHAR): CARDINAL;

    (* From program argument, gets IP address in network byte order. *)

    VAR args: IOChan.ChanId;  L, pos: CARDINAL;  havearg, found: BOOLEAN;
        input: ARRAY [0..79] OF CHAR;

    BEGIN
        args := ArgChan();
        havearg := IsArgPresent();
        IF havearg THEN
            TextIO.ReadString (args, input);
        ELSIF TESTING THEN
            input := testInput;
        ELSE
            input := "";
        END (*IF*);
        WHILE input[0] = ' ' DO Strings.Delete (input, 0, 1) END (*WHILE*);
        L := Strings.Length (input);
        WHILE (L > 0) AND (input[L-1] = ' ') DO
            DEC (L);  input[L] := CHR(0);
        END (*WHILE*);

        (* See whether a username has also been supplied. *)

        username[0] := Nul;
        Strings.FindNext ('@', input, 0, found, pos);
        IF found THEN
            IF pos > 0 THEN
                Strings.Extract (input, 0, pos, username);
            END (*IF*);
            Strings.Delete (input, 0, pos+1);
        END (*IF*);

        RETURN ExtractIPAddress (input);

    END GetParam;

(************************************************************************)
(*                           THE MAIN SERVER CODE                       *)
(************************************************************************)

PROCEDURE ConnectToHost (IPaddress: CARDINAL;  ServerPort: CARDINAL;
                      VAR (*INOUT*) FailureReason: ARRAY OF CHAR): Socket;

    (* Tries to open a connection to the specified host.  Returns the   *)
    (* value NotASocket if we don't succeed; in this case, the string   *)
    (* FailureReason is given a value.                                  *)
    (* IPaddress is in network byte order.                              *)

    VAR s: Socket;  peer: SockAddr;

    BEGIN
        IF IPaddress <> 0 THEN
            s := socket (AF_INET, SOCK_STREAM, AF_UNSPEC);

            IF s = NotASocket THEN
                Strings.Assign ("Can't allocate socket", FailureReason);
            ELSE

                (* Socket open, connect to the server. *)

                WITH peer DO
                    family := AF_INET;
                    WITH in_addr DO
                        port := Swap2(ServerPort);
                        addr := IPaddress;
                        zero := Zero8;
                    END (*WITH*);
                END (*WITH*);

                IF connect (s, peer, SIZE(peer)) THEN

                    Strings.Assign ("Failed to connect", FailureReason);
                    soclose(s);
                    s := NotASocket;

                END (*IF*);
            END (*IF*);

        ELSE

            Strings.Assign ("500 Unknown host", FailureReason);
            s := NotASocket;

        END (*IF*);

        RETURN s;

    END ConnectToHost;

(************************************************************************)

PROCEDURE RunClientSession;

    VAR
        port: IOPort;
        replyQ: PacketQueues.PacketQ;
        ServerAddr: CARDINAL;
        ns: Socket;
        ExceptqActive: BOOLEAN;
        exRegRec: OS2.EXCEPTIONREGISTRATIONRECORD;
        message, username: ARRAY [0..127] OF CHAR;
        ProgVersion: ARRAY [0..31] OF CHAR;

    BEGIN
        (* Initial setup and opening messages. *)

        IF sock_init() <> 0 THEN
            STextIO.WriteString ("No network.");
            STextIO.WriteLn;
            RETURN;
        END (*IF*);

        (* Get server address. *)

        ServerAddr := GetParam(username);
        IF ServerAddr = 0 THEN
            STextIO.WriteString ("Server address? ");
            STextIO.ReadString (message);
            ServerAddr := ExtractIPAddress (message);
            IF ServerAddr = 0 THEN
                STextIO.WriteString ("Invalid address, exiting");  STextIO.WriteLn;
                RETURN;
            END (*IF*);
        END (*IF*);

        ExceptqActive := InstallExceptq (exRegRec);
        IF NOT ExceptqActive THEN
            STextIO.WriteString ("Failed to load the Exceptq handler.");
            STextIO.WriteLn;
        END (*IF*);

        GetProgramName (ProgVersion);
        Strings.Assign (ProgVersion, message);
        STextIO.WriteString (message);  STextIO.WriteLn;

        (* Now connect to the server and run the SSH transport layer. *)

        ns := ConnectToHost (ServerAddr, ServerPort, message);
        IF ns = NotASocket THEN
            STextIO.WriteString (message);
            RETURN;
        END (*IF*);
        replyQ := PacketQueues.CreateQueue();
        port := NewSession (ns, replyQ);
        STextIO.WriteString ("Waiting for key exchange completion");
        STextIO.WriteLn;
        InitWait (port);
        IF NOT DoLogin (port, replyQ, username) THEN
            STextIO.WriteString ("Login failed");  STextIO.WriteLn;
            RETURN;
        END (*IF*);

        (* Main part of this session.  *)

        RunFTPSession (port, replyQ);

        (* Clean up at termination - this is still a problem.  As a workaround  *)
        (* I'm putting in a time delay to allow for packets still in transit.   *)

        Sleep (600);
        ClosePacketSession (port);
        (*PacketQueues.CloseQueue (replyQ);*)

        EVAL (soclose(ns));

        STextIO.WriteString ("End of session");  STextIO.WriteLn;

        UninstallExceptq (exRegRec);

    END RunClientSession;

(********************************************************************************)
(*                                 MAIN PROGRAM                                 *)
(********************************************************************************)

BEGIN
    WeAreServer (FALSE);
    ServerPort := 22;
    RunClientSession;
END sftp.

