IMPLEMENTATION MODULE sshRSA;

        (********************************************************)
        (*                                                      *)
        (*                 ssh-rsa public keys                  *)
        (*             This is the client version               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            24 November 2021                *)
        (*  Last edited:        24 November 2021                *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, ADR, FILL;

IMPORT BigNum, HashAlgs, VarStrings;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS;

FROM BigNum IMPORT
    (* type *)  BN;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransactionL;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL;

(************************************************************************)
(*               EXTRACTING THE COMPONENTS OF A PUBLIC KEY              *)
(************************************************************************)

PROCEDURE DissectPublicKey (VAR (*IN*) key: ByteStr;  VAR (*OUT*) e, n: BN);

    (* Gets the two numbers contained in the key.       *)

    VAR name: ARRAY [0..128] OF CHAR;
        size, j: CARDINAL;

    BEGIN
        size := 0;

        (* Name - although we don't actually use it so far. *)

        Copy (ADR(key.data^[0]), ADR(size), 4);
        size := Swap4 (size);
        Copy (ADR(key.data^[4]), ADR(name), size);
        name[size] := CHR(0);
        j := size + 4;

        (* Exponent (e). *)

        Copy (ADR(key.data^[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        e := BigNum.BinToBN (key.data^, j, size);
        INC (j, size);

        (* Modulus (n). *)

        Copy (ADR(key.data^[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        n := BigNum.BinToBN (key.data^, j, size);

    END DissectPublicKey;

(************************************************************************)
(*                        RETURNING OUR OWN KEY                         *)
(************************************************************************)

PROCEDURE GetFormattedPublicKey (): ByteStr;

    (* A dummy procedure, needed only by server code. *)

    VAR dummy: ByteStr;

    BEGIN
        MakeBS (dummy, 0);
        RETURN dummy;
    END GetFormattedPublicKey;

(************************************************************************)
(*                        GENERATING A SIGNATURE                        *)
(************************************************************************)

PROCEDURE PKCS1_v1_5 (hashf: HashAlgs.HashAlgorithm;
                                    message: ByteStr;  k: CARDINAL): BN;

    (* Implements the EMSA-PKCS1-v1_5-ENCODE operation as specified in  *)
    (* section 9.2 (page 45) of RFC 8017.   This is the initial step    *)
    (* needed in creating a signature.  The result, as a byte string,   *)
    (* must be k bytes long.                                            *)

    (* There is obvious scope here to do the job more efficiently, but  *)
    (* until this is known to be working it's best to stay with the     *)
    (* step-by-step following of the standard.                          *)

    TYPE
        asn1str1 = ARRAY [0..14] OF CARD8;
        asn1str2 = ARRAY [0..18] OF CARD8;

    CONST
        asn1start1 = asn1str1 {30H, 21H, 30H, 09H, 06H, 05H, 2BH, 0EH,
                               03H, 02H, 1AH, 05H, 00H, 04H, 14H};

        asn1start2 = asn1str2 {30H, 31H, 30H, 0DH, 06H, 09H, 60H, 86H,
                               48H, 01H, 65H, 03H, 04H, 02H, 01H, 05H,
                               00H, 04H, 20H};

        asn1start3 = asn1str2 {30H, 51H, 30H, 0DH, 06H, 09H, 60H, 86H,
                               48H, 01H, 65H, 03H, 04H, 02H, 03H, 05H,
                               00H, 04H, 40H};

    VAR j, L, PSLen, tLen: CARDINAL;
        ctx: HashAlgs.HashCtx;
        digest: ARRAY [0..63] OF CARD8;
        T: ARRAY [0..82] OF CARD8;
        EM: ByteStr;
        M: BN;

    BEGIN
        (* Step 1: compute hash of message.  *)

        ctx := HashAlgs.HashInit (hashf);
        HashAlgs.HashUpdate (ctx, message.data^, message.size);
        HashAlgs.HashFinal (ctx, digest);

        (* hLen = 20 to 64 bytes after SHA encoding.    *)
        (* emLen = parameter k = length of RSAKey.n     *)

        (* Step 2: ASN1-encode the digest.  This only requires  *)
        (* appending the hash to the asn1start string.          *)

        CASE hashf OF
            HashAlgs.sha1:
                FOR j := 0 TO 14 DO
                    T[j] := asn1start1[j];
                END (*FOR*);
                j := 15;
                PSLen := k - 38;  tLen := 35;

          | HashAlgs.sha2_256:
                FOR j := 0 TO 18 DO
                    T[j] := asn1start2[j];
                END (*FOR*);
                j := 19;
                PSLen := k - 54;  tLen := 51;

          | HashAlgs.sha2_512:
                FOR j := 0 TO 18 DO
                    T[j] := asn1start3[j];
                END (*FOR*);
                j := 19;
                PSLen := k - 86;  tLen := 83;

        ELSE
                WriteString ("ERROR: unhandled hash function");
                WriteLn;
                HALT;
        END (*CASE*);

        L := HashAlgs.HashLength(hashf);
        Copy (ADR(digest), ADR(T[j]), L);

        (* We require k >= tLen + 11.  The length of PS (see standard)  *)
        (* is k - tLen - 3 bytes, and this must be at least 8.  For the *)
        (* hashing algorithms I am currently using, the numbers are:    *)
        (*                    tLen     PSLen   min k    min bits        *)
        (*   sha1           15+20=35   k-38     46        368           *)
        (*   sha2_256       19+32=51   k-54     62        496           *)
        (*   sha2_512       19+64=83   k-86     94        752           *)

        (* PS = emLen - tLen - 3 bytes of 0FFH. *)
        (* This is k - 38 bytes.                *)

        (* Precede T with some padding. *)

        MakeBS (EM, k);
        EM.data^[0] := 0;
        EM.data^[1] := 1;
        FILL (ADR(EM.data^[2]), 0FFH, PSLen);
        EM.data^[k-tLen-1] := 0;

        (* Move T to the tail of EM.  *)

        Copy (ADR(T), ADR(EM.data^[k-tLen]), tLen);

        (* Convert EM to a Bignum. *)

        M := BigNum.BinToBN (EM.data^, 0, EM.size);
        DiscardBS (EM);
        RETURN M;

    END PKCS1_v1_5;

(************************************************************************)

PROCEDURE Sign (hashf: HashAlgs.HashAlgorithm;  message: ByteStr): ByteStr;

    (* A dummy procedure, needed only by server code. *)

    VAR dummy: ByteStr;

    BEGIN
        MakeBS (dummy, 0);
        RETURN dummy;
    END Sign;

(************************************************************************)

PROCEDURE CheckSignature (hashf: HashAlgs.HashAlgorithm;
                            message, publickey: ByteStr;
                               VAR (*INOUT*) signature: ByteStr;
                                logID: TransactionLogID): BOOLEAN;

    (* Checks that signature is a valid signature of the message,       *)
    (* using the given public key.  The verification method is          *)
    (* specified in RFC 8017 section 8.2.2 (page 37).                   *)
    (* During processing the signature is discarded.                    *)

    CONST bufsize = 2048;

    VAR k, L: CARDINAL;
        S, V, M, key, prime: BN;
        match: BOOLEAN;

    BEGIN
        (* The signature blob starts with an algorithm name.  This is   *)
        (* redundant and may be removed.                                *)

        L := 0;  k := 0;
        Copy (signature.data, ADR(L), 4);
        L := Swap4 (L);
        Copy (ADR(signature.data^[4+L]), ADR(k), 4);
        k := Swap4 (k);
        VarStrings.BSDelete (signature, 0, L+8);

        (* We need to extract two numbers from the public key. *)

        DissectPublicKey (publickey, key, prime);

        (* k should also be the length in bytes of the RSA modulus.  *)

        IF k <> BigNum.NbytesNLZ(prime) THEN
            LogTransactionL (logID, "Wrong signature size");
            RETURN FALSE;
        END (*IF*);

        (* Encrypt the signature using the public key. *)

        S := BigNum.BinToBN (signature.data^, 0, signature.size);
        VarStrings.DiscardBS (signature);
        V := BigNum.ModularPower (S, key, prime);
        BigNum.Discard (key);
        BigNum.Discard (prime);

        (*
        (* WHILE DEBUGGING *)

        (* Dump the message as supplied. *)

        WriteString ("Signature encrypted");  WriteLn;
        BigNum.ToHex (S, buffer, bufsize);
        WriteString (buffer);  WriteLn;

        (* Dump the unencrypted message. *)

        WriteString ("Signature without the encryption");  WriteLn;
        BigNum.ToHex (V, buffer, bufsize);
        WriteString (buffer);  WriteLn;
        (* END DEBUG CODE *)
        *)
        BigNum.Discard (S);

        (* Encode the original message. *)

        M := PKCS1_v1_5 (hashf, message, k);

        (* Now M should be equal to V. *)

        match := BigNum.Cmp (M,V) = 0;

        IF NOT match THEN
            LogTransactionL (logID, "Mismatch of signature content, dumping to debug file");
            (*
            Debug.DebugMes ("Mismatch of signature content.  Signature as supplied:");
            BigNum.ToHex (V, buffer, bufsize);
            Debug.DebugMes (buffer);
            Debug.DebugMes ("Our encoding for comparison:");
            BigNum.ToHex (M, buffer, bufsize);
            Debug.DebugMes (buffer);
            *)
        END (*IF*);

        BigNum.Discard (V);
        BigNum.Discard (M);
        RETURN match;

    END CheckSignature;

(************************************************************************)

END sshRSA.

