(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE sshSession;

        (********************************************************)
        (*                                                      *)
        (*                   Session handler                    *)
        (*                                                      *)
        (*  This is the client version. Although the client and *)
        (*  server code are almost identical, we separate them  *)
        (*  out because we need more legality checks in the     *)
        (*  server case.                                        *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            4 January 2021                  *)
        (*  Last edited:        13 January 2022                 *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT OS2, PacketQueues, Strings;

FROM SYSTEM IMPORT
    (* type *)  ADDRESS,
    (* proc *)  CAST;

FROM TransportLayer IMPORT
    (* proc *)  RunTransportLayer;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  OpenPacketSession, ClosePacketSession;

FROM Sockets IMPORT
    (* type *)  Socket, SockAddr,
    (* proc *)  getpeername, soclose;

FROM Inet2Misc IMPORT
    (* proc *)  IPToString;

FROM Conversions IMPORT
    (* proc *)  CardinalToString;

FROM LogCtx IMPORT
    (* var  *)  sshctx;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  CreateLogID, DiscardLogID, LogTransaction;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, DestroyLock, Obtain, Release, CreateTask1;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM MiscFuncs IMPORT
    (* proc *)  EVAL;

(********************************************************************************)

CONST NILQ = CAST(PacketQueues.PacketQ, NIL);

TYPE
    (* Data used in creating a new instance of the session handler task. *)

    NewSessionPointer = POINTER TO NewSessionRecord;
    NewSessionRecord = RECORD
                           port: IOPort;
                           LogID: TransactionLogID;
                           Q: PacketQueues.PacketQ;
                       END (*RECORD*);

VAR
    (* Lock to slow down excessive thread creation during a denial      *)
    (* of service attack.                                               *)

    ThreadCreationLock: Lock;

(********************************************************************************)
(*                                RUN ONE SESSION                               *)
(********************************************************************************)

PROCEDURE RunSession (arg: ADDRESS);

    (* Runs as a separate thread, keeps looping until told to exit. *)

    VAR NSP: NewSessionPointer;
        sess: IOPort;
        replyQ: PacketQueues.PacketQ;
        LogID: TransactionLogID;

    BEGIN
        OS2.DosError (OS2.FERR_DISABLEHARDERR);              (* disable hard error popups *)

        (* Copy the NewSessionPointer^ information. *)

        NSP := arg;

        Obtain (ThreadCreationLock);
        sess := NSP^.port;
        LogID := NSP^.LogID;
        replyQ := NSP^.Q;
        Release (ThreadCreationLock);
        DEALLOCATE (NSP, SIZE(NewSessionRecord));

        (* Now we start talking to the peer.  *)

        RunTransportLayer (sess, replyQ, NIL, LogID);

        (*
        IF replyQ <> NILQ THEN
            PacketQueues.CloseQueue (replyQ);
        END (*IF*);
        *)
        DiscardLogID (LogID);

    END RunSession;

(********************************************************************************)
(*                            STARTING A NEW SESSION                            *)
(********************************************************************************)

PROCEDURE NewSession (S: Socket;  replyQ: PacketQueues.PacketQ): IOPort;

    (* Launches a session. *)

    (* Starts and runs one instance of the SSH transport layer.  The session    *)
    (* runs in a separate thread; this procedure returns after starting the     *)
    (* session, it does not wait until the session is over.                     *)

    VAR NSP: NewSessionPointer;
        ourport: IOPort;
        LogFilePrefix: ARRAY [0..6] OF CHAR;
        ID: TransactionLogID;
        success: BOOLEAN;

    BEGIN
        Obtain (ThreadCreationLock);

        (* Create the log file ID, open the encrypted port.  Although we don't  *)
        (* do any logging in the client code, we need a log ID because the      *)
        (* transport layer uses common code for both server and client.         *)

        LogFilePrefix := "       ";
        ID := CreateLogID (sshctx, LogFilePrefix);
        ourport := OpenPacketSession (S, FALSE, ID);

        NEW (NSP);
        WITH NSP^ DO
            LogID := ID;
            port := ourport;
            Q := replyQ;
        END (*WITH*);
        success := CreateTask1 (RunSession, 3, "ssh session", NSP);
        Release (ThreadCreationLock);
        IF NOT success THEN
            ClosePacketSession (ourport);
            soclose (S);
            DEALLOCATE (NSP, SIZE(NewSessionRecord));
            RETURN NIL;
        END (*IF*);
        RETURN ourport;
    END NewSession;

(********************************************************************************)
(*                            MODULE INITIALISATION                             *)
(********************************************************************************)

BEGIN
    CreateLock (ThreadCreationLock);
FINALLY
    DestroyLock (ThreadCreationLock);
END sshSession.

